
__author__    = "Andre Merzky, Mark Santcroos"
__copyright__ = "Copyright 2014-2015, The SAGA Project"
__license__   = "MIT"


""" (GSI)SSH based Globus Online Adaptor """

import os
import threading


from ...              import exceptions as rse
from ...url           import Url
from ...utils         import pty_shell  as rsups
from ...utils         import misc       as rsumisc
from ...              import filesystem as api
from ...adaptors      import base       as a_base
from ...adaptors.cpi  import filesystem as cpi
from ...adaptors.cpi  import decorators as cpi_decs


SYNC_CALL  = cpi_decs.SYNC_CALL
ASYNC_CALL = cpi_decs.ASYNC_CALL


# TODO: We could make this configurable,
# so people without gsissh can still perform some operations,
# as they could activate the endpoints through the globus web interface.
# go+ssh:// vs go+gsissh:// comes to mind
GO_DEFAULT_URL = "gsissh://cli.globusonline.org/"


# ------------------------------------------------------------------------------
# the adaptor name
#
_ADAPTOR_NAME          = "radical.saga.adaptors.globus_online_file"
_ADAPTOR_SCHEMAS       = ["go"]  # TODO: also allow file:// ??

# ------------------------------------------------------------------------------
# the adaptor capabilities & supported attributes
#
_ADAPTOR_CAPABILITIES  = {
    "metrics"          : [],
    "contexts"         : {"x509"     : "X509 proxy for Globus",
                          "userpass" : "username/password for GlobusOnline"}
}

# ------------------------------------------------------------------------------
# the adaptor documentation
#
_ADAPTOR_DOC           = {
    "name"             : _ADAPTOR_NAME,
    "capabilities"     : _ADAPTOR_CAPABILITIES,
    "description"      : """
        The globusonline file adaptor. This adaptor uses the GO file transfer
        service (https://www.globus.org/).
        """,
    "details"          : """
        """,
    "schemas"          : {"go": "use globus online for gridftp file transfer"}
}

# ------------------------------------------------------------------------------
# the adaptor info is used to register the adaptor with SAGA

_ADAPTOR_INFO          = {
    "name"             : _ADAPTOR_NAME,
    "version"          : "v0.2",
    "schemas"          : _ADAPTOR_SCHEMAS,
    "cpis"             : [
                             {
                                 "type"  : "radical.saga.namespace.Directory",
                                 "class" : "GODirectory"
                             },
                             {
                                 "type"  : "radical.saga.namespace.Entry",
                                 "class" : "GOFile"
                             },
                             {
                                 "type"  : "radical.saga.filesystem.Directory",
                                 "class" : "GODirectory"
                             },
                             {
                                 "type"  : "radical.saga.filesystem.File",
                                 "class" : "GOFile"
                             }
                         ]
}


################################################################################
# The adaptor class
class Adaptor(a_base.Base):
    """
    This is the actual adaptor class, which gets loaded by SAGA (i.e. by the
    SAGA engine), and which registers the CPI implementation classes which
    provide the adaptor's functionality.
    """

    # --------------------------------------------------------------------------
    #
    def __init__(self):

        a_base.Base.__init__(self, _ADAPTOR_INFO)

        self.notify       = self._cfg['enable_notifications']
        self.f_mode       = self._cfg['failure_mode']
        self.prompt       = self._cfg['prompt_pattern']
        self.localhost_ep = self._cfg['localhost_endpoint']
        self.shells       = dict()  # keep go shells for each session

        #
        # Lock to synchronize concurrent access to data structures
        #
        self.shell_lock = threading.RLock()

    # --------------------------------------------------------------------------
    #
    def sanity_check(self):
        pass

    # --------------------------------------------------------------------------
    #
    def get_go_shell(self, session, go_url=GO_DEFAULT_URL):

        # This returns a pty shell for: '[gsi]ssh username@cli.globusonline.org'
        #
        # X509 contexts are preferred, but ssh contexts, userpass and myproxy
        # can also be used.  If the given url has username / password encoded,
        # we create an userpass context out of it and add it to the (copy of)
        # the session.

        self._logger.debug("Acquiring lock")
        with self.shell_lock:
            self._logger.debug("Acquired lock")

            sid = session._id

            init = False
            create = False

            if sid in self.shells \
                and self.shells[sid]['shell'].alive(recover=False):

                self._logger.debug("Shell in cache and alive, can reuse.")

            elif sid in self.shells:
                self._logger.debug("Shell in cache but not alive.")
                self.shells[sid]['shell'].finalize()
                self._logger.debug("Shell is finalized, need to recreate.")
                create = True

            else:
                self._logger.debug("Shell not in cache, create entry.")
                init = True
                create = True
                self.shells[sid] = {}

            # Acquire new shell
            if create:

                # deep copy URL (because of?)
                new_url = Url(go_url)

                # GO specific prompt pattern
                opts = {'prompt_pattern': self.prompt}

                # create the shell.
                shell = rsups.PTYShell(new_url, session=session,
                                    logger=self._logger, opts=opts, posix=False)
                self.shells[sid]['shell'] = shell

                # For this fresh shell, we get the list of public endpoints.
                # That list will contain the set of hosts we can potentially
                # connect to.
                self.get_go_endpoint_list(session, shell, fetch=True)

            # Initialize other dict members and remote shell
            if init:
                shell = self.shells[sid]['shell']

                # Confirm the user ID for this shell
                self.shells[sid]['user'] = None
                _, out, _ = shell.run_sync('profile')
                for line in out.split('\n'):
                    if 'User Name:' in line:
                        self.shells[sid]['user'] = line.split(':', 2)[1].strip()
                        self._logger.debug("using account '%s'"
                                          % self.shells[sid]['user'])
                        break

                if not self.shells[sid]['user']:
                    raise rse.NoSuccess("Could not confirm user id")

                # Toggle notification
                if self.notify == 'True':
                    self._logger.debug("enable email notifications")
                    shell.run_sync('profile -n on')
                elif self.notify == 'False':
                    self._logger.debug("disable email notifications")
                    shell.run_sync('profile -n off')

            self._logger.debug("Release lock")

            # we have the shell for sure by now -- return it!
            return self.shells[sid]['shell']

    # --------------------------------------------------------------------------
    #
    # Constructs the following from a SAGA/user URL:
    #
    #   - host: the original host element from the input URL
    #   - name: the EP name as used to interact with GO
    #   - url: the URL corrected with the EP name as the host element
    #
    def get_go_endpoint_ids(self, session, url):

        sid = session._id

        if sid not in self.shells:
            raise rse.IncorrectState("GO shell disconnected")

        ep_url = Url()
        ep_url.schema = url.schema
        ep_url.port = url.port

        if '#' in url.host:
            # We assume an already existing EP.
            # We use the name for the host entry also.
            ep_name = url.host
            ep_url.host = ep_name
        elif url.host == 'localhost':
            # Special case to translate "localhost" to the users own machine.
            if self.localhost_ep != 'None':
                ep_name = self.localhost_ep
                ep_url.host = ep_name
            else:
                # TODO: could add more heuristics, like:
                # - looking for a "username#localhost"
                # - looking for special type of entries in the endpoint-list
                raise rse.BadParameter("localhost endpoint not configured")
        else:
            # Create an EP based on the username and hostname
            ep_name = "%s#%s" % (self.shells[sid]['user'], url.host)
            ep_url.host = url.host

        return ep_name, ep_url


    # --------------------------------------------------------------------------
    #
    def get_path_spec(self, session, url, path=None, cwd_url=None,
                      cwd_path=None):

        # we assume that, whenever we request a path spec, we also want to use
        # it, and we thus register and activate the endpoint, if needed.

        sid = session._id

        if sid not in self.shells:
            raise rse.IncorrectState("GO shell disconnected")

        shell = self.shells[sid]['shell']
        url   = Url(url)

        if not path:
            path = url.path

        if not cwd_url:
            cwd_url = Url(url)

            if not cwd_path:
                cwd_path = '.'
        else:
            if not cwd_path:
                cwd_path = cwd_url.path

        if not url.host:
            url.host = cwd_url.host
        if not url.schema:
            url.schema = cwd_url.schema

        if not url.host:
            raise rse.BadParameter('need host for GO ops')

        if not url.schema:
            raise rse.BadParameter('need schema for GO ops')

        ep_name, ep_url = self.get_go_endpoint_ids(session, url)

        # if both URLs point into the same namespace, and if the given path is
        # not absolute, then expand it relative to the cwd_path (if it exists).
        # Otherwise it is left to the unmodified path.
        ps_path = path
        # TODO: should the check be on cwd_url / url or on ep_url?
        if rsumisc.url_is_compatible(cwd_url, url) and not path.startswith('/'):
            if cwd_path and path:
                ps_path = os.path.join(cwd_path, path)
            elif cwd_path:
                ps_path = cwd_path

        # the pathspec is the concatenation of ps_host and ps_path
        ps = "%s%s" % (ep_name, ps_path)

        # check if we know the endpoint in XXX, and create/activate as needed
        self.get_go_endpoint(session, shell, ep_url)

        return ps

    # --------------------------------------------------------------------------
    #
    def get_go_endpoint(self, session, shell, url):

        # for the given URL, derive the endpoint string.
        ep_name, ep_url = self.get_go_endpoint_ids(session, url)

        ep = self.get_go_endpoint_list(session, shell, ep_name, fetch=False)

        if not ep:

            # if not, check if it was created meanwhile (fetch again)
            ep = self.get_go_endpoint_list(session, shell, ep_name, fetch=True)

            if not ep:

                # Don't try to create endpoints that are supposed to be there
                # (plus, we can't!)
                if '#' in url.host:
                    raise rse.NoSuccess("# in hostname, not going to create!")

                # if not, create it, activate it, and refresh all entries
                shell.run_sync("endpoint-add %s -p %s" % (ep_name, ep_url))

                # refresh endpoint entries again
                ep = self.get_go_endpoint_list(session, shell, ep_name,
                                               fetch=True)

                if not ep:
                    # something above must have failed ...
                    raise rse.NoSuccess("endpoint initialization failed")

        # we have the endpoint now, for sure -- make sure its activated
        if not ep['Credential Status'] == 'ACTIVE':

            # TODO: I had an active endpoint, but still got an activation
            #       prompt, probably because the remaining lifetime was not
            #       very long anymore.
            # or Credential Time Left    : 00:16:35 < ????
            # Answer: below 30 min there is a activation prompt,
            # but it does actually continue normally.
            # Need to capture that behavior.

            # Only Globus Connect Service Endpoints don't need -g?
            # Had contact on this with Globus Support, they couldn't suggest
            # anything better.
            if ep['MyProxy Server'] == 'myproxy.globusonline.org' and \
                     '/C=US/O=Globus Consortium/OU=Globus Connect Service/CN=' \
                     in ep['Credential Subject']:
                shell.run_sync("endpoint-activate %s" % ep_name)
            else:
                shell.run_sync("endpoint-activate -g %s" % ep_name)

            # reload list to check status
            ep = self.get_go_endpoint_list(session, shell, ep_name, fetch=True)

            if not ep['Credential Status'] == 'ACTIVE':
                raise rse.AuthorizationFailed("endpoint activation failed")

        return ep


    # --------------------------------------------------------------------------
    #
    def get_go_endpoint_list(self, session, shell, ep_name=None, fetch=False):

        # if 'fetch' is True, query the shell for an updated endpoint list.
        # then check if the given ep_name is a known endpoint name, and if so,
        # return that entry -- otherwise return None.  If no ep_name is given,
        # and fetch is True, we thus simply refresh the internal list.

        if fetch:

            with self.shell_lock:

                endpoints = {}
                name = None

                if ep_name:
                    endpoint_selection = ep_name
                else:
                    endpoint_selection = '-a'

                # Get the details of endpoints _OWNED_ by user
                _, out, _ = shell.run_sync(
                                     "endpoint-details %s -f "
                                     "legacy_name,"         # Legacy Name
                                     "credential_status,"   # Credential Status
                                     "credential_subject,"  # Credential Subject
                                     "myproxy_server"       # MyProxy Server
                                     % endpoint_selection
                                    )

                for line in out.split('\n'):
                    elems = line.split(':', 1)

                    if len(elems) != 2:
                        continue

                    key = elems[0].strip()
                    val = elems[1].strip()

                    if not key or not val:
                        continue

                    if key == "Legacy Name":

                        # we now operate on a new entry -- initialize it
                        name = val

                        endpoints[name] = {}
                        endpoints[name]['Legacy Name']       = name

                    else:

                        # Continued passed of an entry, name should always exist
                        try:
                            endpoints[name][key] = val
                        except Exception as e:
                            raise rse.NoSuccess("No entry to operate on: %s[%s]"
                                    % (key,val)) from e

                # replace the ep info dist with the new one, to clean out old
                # entries.
                # TODO: merge and not replace(?)
                #
                self.shells[session._id]['endpoints'] = endpoints

        if ep_name:
            # return the requested entry, or None
            return self.shells[session._id]['endpoints'].get(ep_name, None)


    # --------------------------------------------------------------------------
    #
    def run_go_cmd(self, shell, cmd, mode=None):

        # available modes:
        #   raise : raise NoSuccess on error
        #   report: print error message, but continue
        #   ignore: do nothing

        if not mode:
            mode = self.f_mode

        _, out, err = shell.run_sync(cmd)

        # see if the second line starts with 'Error'.  Note that this assumes
        # that the command only is one line...
        lines = out.split('\n')

        if len(lines) > 1:
            if lines[1].startswith('Error:'):
                err = "%s\n%s" % (err, '\n'.join(lines))
                out = None

            else:
                # on success, we always remove the first line, as that is not
                # part of the output, really (this shell does not support
                # 'stty -echo'...
                out = '\n'.join(lines[1:])

        if err:
            if mode == 'raise':
                if lines[3].startswith('Message:'):
                    cause = lines[3].split(':')[1].strip()
                    if cause == 'No such file or directory':
                        raise rse.DoesNotExist(cause)
                    elif cause == 'Could not connect to server':
                        raise rse.BadParameter(cause)
                    else:
                        raise rse.NoSuccess('Unknown error: %s' % cause)
                else:
                    raise rse.NoSuccess('Could not parse error: %s' % err)

                # TODO: Handle GO access to directories that are not allowed
                # TODO: by Globus Personal by default (e.g. /var/ , /tmp)
                # TODO: '''Message: Fatal FTP Response
                # TODO:    ---
                # TODO:    500 Command failed : Path not allowed.'''

            if mode == 'report':
                self._logger.error("Error in '%s': %s" % (cmd, err))

            if mode == 'silent':
                pass

        return out, err

    # --------------------------------------------------------------------------
    #
    def mkparents(self, shell, tgt_ps):

        # GO does not support mkdir -p, so we need to split the dir into
        # elements and create one after the other, ignoring errors for already
        # existing elements.
        # TODO: Can't we check for existence? The errors are confusing.

        host_ps, path_ps = tgt_ps.split('/', 1)
        path_ps = '/' + path_ps

        self._logger.info('mkparents %s' % path_ps)

        if path_ps.startswith('/'):
            cur_path = ''
        else:
            cur_path = '.'

        error = {}
        path_elems = [_f for _f in path_ps.split('/') if _f]

        for path_elem in path_elems :

            cur_path = "%s/%s" % (cur_path, path_elem)
            out, err = self.run_go_cmd(shell, "mkdir %s%s"
                                              % (host_ps, cur_path))

            if err:
                error[cur_path] = err

        if len(error):

            # some mkdir gave an error.  Check if the error occurred on the last
            # dir (the tgt), and if that is not a ignorable report that it
            # already exists -- anything else will raise an exception though...
            if cur_path in error:

                if 'Path already exists' not in error[cur_path]:

                    if self.f_mode == 'raise':
                        # TODO: 'translate_exception' call would be useful here
                        # TODO: We can use the exceptions from run_go_cmd?
                        raise rse.NoSuccess("Could not make dir hierarchy: %s"
                                           % str(error))

                    if self.f_mode == 'report':
                        self._logger.error("Could not make dir hierarchy: %s"
                                          % str(error))

                    if self.f_mode == 'silent':
                        pass

    def go_transfer(self, shell, flags, source, target):

        # TODO: I dont think we handle relative targets yet

        self._logger.debug('Adaptor:go_transfer(%s, %s)' % (source, target))

        # 0: Copy files that do not exist at the destination
        sync_level = 0

        # Create parents
        if flags & api.CREATE_PARENTS:
            self.mkparents(shell, os.path.dirname(target))

      # if flags & api.OVERWRITE:
            # 1: Copy files if the size of the destination does not match the
            #    size of the source
            # 2: Copy files if the timestamp of the destination is older than
            #    the timestamp of the source
            # 3: Copy files if checksums of the source and destination do not
            #    match
      #    sync_level = 3

        # Set recursive flag
        cmd_flags = ""
        if flags & api.RECURSIVE:
            cmd_flags += "-r"

        # Initiate background copy
        # TODO: Should we use a deadline?
        cmd = "transfer %s -s %d -- '%s' '%s'" \
            % (cmd_flags, sync_level, source, target)

        out, _ = self.run_go_cmd(shell, cmd)
        # 'Task ID: 8c6f989d-b6aa-11e4-adc6-22000a97197b'
        key, value = out.split(':')
        if key != 'Task ID':
            raise Exception("Expected Task ID: <id>, got %s" % out)
        task_id = value.strip()

        # Wait until background copy has finished
        cmd = "wait -q %s" % task_id
        self.run_go_cmd(shell, cmd)

        # Retrieve task status
        cmd = "status -f status %s" % task_id
        out, _ = self.run_go_cmd(shell, cmd)
        # Status: SUCCEEDED
        key, value = out.split(':')
        if key != 'Status':
            raise Exception("Expected Status: <status>, got %s" % out)
        status = value.strip()

        # Validate task status
        if status == 'SUCCEEDED':
            # The task completed successfully.
            return

        elif status == 'ACTIVE':
            # The task is in progress.
            raise Exception('Task active, this should not happen after wait')

        elif status == 'INACTIVE':
            # The task has been suspended and will not continue without
            # intervention.  Currently, only credential expiration will cause
            # this state.
            raise Exception('Task inactive, probably credentials have expired')

        elif status == 'FAILED':
            # The task or one of its subtasks failed, expired, or was canceled.
            raise Exception('Task failed')

        else:
            raise Exception('Unknown status: %s' % status)


    # --------------------------------------------------------------------------
    #
    # Helper function to test for existence and type.
    # Will raise DoesNotExist for non-existing entries.
    def stat(self, shell, ps):

        # TODO: stat() would probably benefit from some caching

        out, err = self.run_go_cmd(shell, "ls -la '%s'" % ps, mode='raise')

        mode = out.split('\n')[0].split()[0][0]
        if   mode == '-': mode = 'file'
        elif mode == 'd': mode = 'dir'
        elif mode == 'l': mode = 'link'
        else: raise rse.NoSuccess("stat unknown mode: '%s' (%s)" % (mode, out))

        size = int(out.split('\n')[0].split()[3])

        return {'mode': mode,
                'size': size }


################################################################################
#
class GODirectory(cpi.Directory):
    """ Implements cpi.Directory """

    # --------------------------------------------------------------------------
    #
    def __init__(self, api, adaptor):

        _cpi_base = super(GODirectory, self)
        _cpi_base.__init__(api, adaptor)

    # --------------------------------------------------------------------------
    #
    def _is_valid(self):

        if not self.valid:
            raise rse.IncorrectState("this instance was closed or removed")

    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def init_instance(self, adaptor_state, url, flags, session):
        """ Directory instance constructor """

        # TODO: eval flags!
        if flags is None:
            flags = 0

        self.orig     = Url(url)  # deep copy
        self.url      = Url(url)  # deep copy
        self.path     = url.path  # keep path separate
        self.url.path = None

        self.flags    = flags
        self.session  = session
        self.valid    = False     # will be set by initialize

        self.initialize()

        return self.get_api()

    # --------------------------------------------------------------------------
    #
    def initialize(self):

        # GO shell got started, found its prompt.  Now, change
        # to the initial (or later current) working directory.

        self.shell = self._adaptor.get_go_shell(self.session)
        self.ep    = self._adaptor.get_go_endpoint(self.session, self.shell,
                                                   self.url)

        ps = self.get_path_spec()

        if not self.ep:
            raise rse.BadParameter("invalid dir '%s'" % ps)

        if self.flags & api.CREATE_PARENTS:
            self._adaptor.mkparents(self.shell, ps)

        elif self.flags & api.CREATE:
            # TODO: check for errors?
            self._adaptor.run_go_cmd(self.shell, "mkdir '%s'" % ps)

        else:
            stat = self._adaptor.stat(self.shell, ps)
            if stat['mode'] not in ['dir', 'link']:
                # TODO: if link, check the target
                raise rse.IncorrectState('Is not a directory')

        self._logger.debug("Init directory %s/%s" % (self.url, self.path))

        self.valid = True

    # --------------------------------------------------------------------------
    #
    def finalize(self, kill=False):

        if kill and self.shell:
            self.shell.finalize(True)
            self.shell = None

        self.valid = False

    # --------------------------------------------------------------------------
    #
    def get_path_spec(self, url=None, path=None):

        return self._adaptor.get_path_spec(session=self.session,
                                           url=url,
                                           path=path,
                                           cwd_url=self.url,
                                           cwd_path=self.path)

    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def open(self, url, flags):

        self._is_valid()

        adaptor_state = {"from_open" : True,
                         "url"       : Url(self.url),   # deep copy
                         "path"      : self.path}

        if rsumisc.url_is_relative(url):
            url = rsumisc.url_make_absolute(self.get_url(), url)

        return api.File(url=url, flags=flags, session=self.session,
                           _adaptor=self._adaptor, _adaptor_state=adaptor_state)

    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def open_dir(self, url, flags):

        self._is_valid()

        adaptor_state = {"from_open": True,
                         "url"      : Url(self.url),   # deep copy
                         "path"     : self.path}

        if rsumisc.url_is_relative(url):
            url = rsumisc.url_make_absolute(self.get_url(), url)

        return api.Directory(url=url, flags=flags, session=self.session,
                _adaptor=self._adaptor, _adaptor_state=adaptor_state)

    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def change_dir(self, tgt, flags):

        tgt_url = Url(tgt)

        # TODO: attempt to get new EP
        if not rsumisc.url_is_compatible(self.url, tgt_url):
            raise rse.BadParameter("Target dir outside of namespace '%s': %s"
                                  % (self.url, tgt_url))

        if rsumisc.url_is_relative(tgt_url):

            self.path      = tgt_url.path
            self.orig.path = self.path

        else:
            self.orig      = Url(tgt_url)
            self.url       = tgt_url
            self.path      = self.url.path
            self.url.path  = None

        self.initialize()

        self._logger.debug("changed directory (%s)" % (tgt))

    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def close(self, timeout=None):

        if timeout:
            raise rse.BadParameter("timeout for close not supported")

        self.finalize(kill=True)

    # --------------------------------------------------------------------------
    @SYNC_CALL
    def get_url(self):

        self._is_valid()

        return Url(self.orig)  # deep copy

    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def list(self, npat, flags):

        self._is_valid()

        npat_ps  = self.get_path_spec(url=npat)
        # TODO: catch errors?
        out, err = self._adaptor.run_go_cmd(self.shell, "ls '%s'" % (npat_ps))
        lines = [_f for _f in out.split("\n") if _f]
        self._logger.debug(lines)

        self.entries = []
        for line in lines:
            self.entries.append(Url(line.strip()))

        return self.entries


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def copy_self(self, tgt, flags):

        self._is_valid()

        self._logger.debug('Directory:copy_self(%s)' % tgt)

        return self.copy(src_in=None, tgt_in=tgt, flags=flags)

    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def copy(self, src_in, tgt_in, flags, _from_task=None):

        self._logger.debug('Directory:copy(%s, %s)' % (src_in, tgt_in))
        self._is_valid()

        src_ps = self.get_path_spec(url=src_in)
        tgt_ps = self.get_path_spec(url=tgt_in)

        # Check for existence of source
        self._adaptor.stat(self.shell, src_ps)

        self._adaptor.go_transfer(self.shell, flags, src_ps, tgt_ps)

    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def move_self(self, tgt, flags):

        return self.move(src_in=None, tgt_in=tgt, flags=flags)


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def move(self, src_in, tgt_in, flags):

        # if src and target are on the same endpoint, we might get away with an
        # actual 'rename' command -- in all other cases (or if rename failed),
        # we fake move as non-atomic copy/remove...

        src_ps = self.get_path_spec(url=src_in)
        tgt_ps = self.get_path_spec(url=tgt_in)

        src_ep_str = src_ps.split('/', 1)[0]
        tgt_ep_str = tgt_ps.split('/', 1)[0]

        if src_ep_str == tgt_ep_str:

            try:
                self._adaptor.run_go_cmd(self.shell, "rename '%s' '%s'"
                                        % (src_ps, tgt_ps))
                return
            except:
                self._logger.warning("rename op failed -- retry as copy/remove")

        # either the op spans endpoints, or the 'rename' op failed
        self.copy(src_in, tgt_in, flags)
        self.remove(src_in, flags)


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def remove_self(self, flags):

        self._is_valid()

        self.remove(self.url, flags)
        self.invalid = True


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def remove(self, tgt_in, flags):

        self._is_valid()

        tgt_ps = self.get_path_spec(url=tgt_in)

        cmd_flags = ""
        if flags & api.RECURSIVE:
            cmd_flags += "-r"

        # TODO: check for errors

        # TODO: a dir only gets removed (on some endpoints)
        # if the trailing '/' is specified -- otherwise the op *silently fails*!
        # Oh well, since we don't really (want to) know if the target is a dir
        # or not, we remove both versions... :/
        cmd      = "rm %s -f '%s/'" % (cmd_flags, tgt_ps)
        out, err = self._adaptor.run_go_cmd(self.shell, cmd)

        cmd      = "rm %s -f '%s'"  % (cmd_flags, tgt_ps)
        out, err = self._adaptor.run_go_cmd(self.shell, cmd, mode='ignore')


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def make_dir(self, tgt_in, flags):

        self._is_valid()

        tgt_ps = self.get_path_spec(url=tgt_in)

        # TODO: check for errors

        if flags & api.CREATE_PARENTS:
            self._adaptor.mkparents(self.shell, tgt_ps)

        else:
            cmd = "mkdir '%s'" % tgt_ps
            self._adaptor.run_go_cmd(self.shell, cmd)


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def get_size_self(self):

        self._is_valid()

        return self.get_size(self.url)


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def get_size(self, tgt_in):

        self._is_valid()

        tgt_ps = self.get_path_spec(url=tgt_in)

        stat = self._adaptor.stat(self.shell, tgt_ps)

        return stat['size']

    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def is_dir_self(self):

        self._is_valid()

        return self.is_dir(self.url)

    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def is_dir(self, tgt_in):

        self._is_valid()

        tgt_ps = self.get_path_spec(url=tgt_in)

        stat = self._adaptor.stat(self.shell, tgt_ps)

        if stat['mode'] == 'dir':
            return True
        else:
            return False

    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def is_link_self(self):

        self._is_valid()

        return self.is_link(self.url)

    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def is_link(self, tgt_in):

        self._is_valid()

        tgt_ps = self.get_path_spec(url=tgt_in)

        stat = self._adaptor.stat(self.shell, tgt_ps)

        if stat['mode'] == 'link':
            return True
        else:
            return False

    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def is_file_self(self):

        self._is_valid()

        return self.is_link(self.url)

    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def is_file(self, tgt_in):

        self._is_valid()

        tgt_ps = self.get_path_spec(url=tgt_in)

        stat = self._adaptor.stat(self.shell, tgt_ps)

        if stat['mode'] == 'file':
            return True
        else:
            return False

    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def exists(self, tgt_in):

        self._is_valid()

        tgt_ps = self.get_path_spec(url=tgt_in)

        try:
            self._adaptor.stat(self.shell, tgt_ps)
        except rse.DoesNotExist:
            return False

        return True


###############################################################################
#
class GOFile(cpi.File):
    """ Implements cpi.File
    """
    # --------------------------------------------------------------------------
    #
    def __init__(self, api, adaptor):

        _cpi_base = super(GOFile, self)
        _cpi_base.__init__(api, adaptor)

    # --------------------------------------------------------------------------
    #
    def _is_valid(self):

        if not self.valid:
            raise rse.IncorrectState("this instance was closed or removed")

    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def init_instance(self, adaptor_state, url, flags, session):
        """ File instance constructor """

        # TODO: eval flags!
        if flags is None:
            flags = 0

        self.orig     = Url(url)  # deep copy
        self.url      = Url(url)  # deep copy
        self.path     = url.path  # keep path separate
        self.cwd      = rsumisc.url_get_dirname(self.url)
        self.url.path = None

        self.flags    = flags
        self.session  = session
        self.valid    = False     # will be set by initialize

        self.initialize()

        return self.get_api()

    # --------------------------------------------------------------------------
    #
    def initialize(self):

        # GO shell got started, found its prompt.  Now, change
        # to the initial (or later current) working directory.

        self.shell = self._adaptor.get_go_shell(self.session)
        self.ep    = self._adaptor.get_go_endpoint(self.session, self.shell,
                                                   self.url)
        ps         = self.get_path_spec()
        cwd_ps     = self.get_path_spec(path=self.cwd)

        if not self.ep:
            raise rse.BadParameter("invalid file '%s'" % ps)

        if self.flags & api.CREATE_PARENTS:
            self._adaptor.mkparents(self.shell, cwd_ps)

        elif self.flags & api.CREATE:
            self._logger.error("CREATE not supported for files via globus online")

        else:
            stat = self._adaptor.stat(self.shell, ps)
            if stat['mode'] not in ['file', 'link']:
                # TODO: if link, check the target
                raise rse.IncorrectState('Is not a (regular) file')

        self._logger.debug("Initialized file %s/%s" % (self.url, self.path))

        self.valid = True


    # --------------------------------------------------------------------------
    #
    def get_path_spec(self, url=None, path=None):

        return self._adaptor.get_path_spec(session=self.session,
                                           url=url,
                                           path=path,
                                           cwd_url=self.url,
                                           cwd_path=self.path)


    # --------------------------------------------------------------------------
    #
    def finalize(self, kill=False):

        if kill and self.shell:
            self.shell.finalize(True)
            self.shell = None

        self.valid = False


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def close(self, timeout=None):

        if timeout:
            raise rse.BadParameter("timeout for close not supported")

        self.finalize(kill=True)

    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def get_url(self):

        self._is_valid()

        return Url(self.orig)  # deep copy

    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def copy_self(self, tgt_in, flags):

        self._is_valid()

        src_ps = self.get_path_spec()
        tgt_ps = self.get_path_spec(url=tgt_in)

        self._adaptor.go_transfer(self.shell, flags, src_ps, tgt_ps)

    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def move_self(self, tgt_in, flags):

        # if src and target are on the same endpoint, we might get away with an
        # actual 'rename' command -- in all other cases (or if rename failed),
        # we fake move as non-atomic copy/remove...

        src_ps = self.get_path_spec()
        tgt_ps = self.get_path_spec(url=tgt_in)

        src_ep_str = src_ps.split('/', 1)[0]
        tgt_ep_str = tgt_ps.split('/', 1)[0]

        # TODO: check for errors

        # TODO: should me extracted and merged with operations on Directory?

        if src_ep_str == tgt_ep_str:

            try:
                self._adaptor.run_go_cmd(self.shell, "rename '%s' '%s'"
                                        % (src_ps, tgt_ps))
                return
            except:
                self._logger.warning("Rename op failed -- retry as copy/remove")

        # either the op spans endpoints, or the 'rename' op failed
        self.copy_self(tgt_in, flags)
        self.remove_self(flags)

    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def remove_self(self, flags):

        self._is_valid()

        cmd_flags = ""
        if flags & api.RECURSIVE:
            cmd_flags += "-r"

        # TODO: check for errors
        # TODO: should me extracted and merged with operations on Directory?

        tgt_ps   = self.get_path_spec()
        cmd      = "rm %s -f '%s'"  % (cmd_flags, tgt_ps)
        out, err = self._adaptor.run_go_cmd(self.shell, cmd, mode='ignore')

    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def get_size_self(self):

        self._is_valid()

        tgt_ps = self.get_path_spec(url=self.url)

        stat = self._adaptor.stat(self.shell, tgt_ps)

        return stat['size']

    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def is_link_self(self):
        self._is_valid()

        tgt_ps = self.get_path_spec(url=self.url)

        stat = self._adaptor.stat(self.shell, tgt_ps)
        if stat['mode'] == 'link':
            return True
        else:
            return False

    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def is_file_self(self):
        self._is_valid()

        tgt_ps = self.get_path_spec(url=self.url)

        stat = self._adaptor.stat(self.shell, tgt_ps)
        if stat['mode'] == 'file':
            return True
        else:
            return False

    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def is_dir_self(self):

        # TODO: can be further extracted and merged with Directory calls

        self._is_valid()

        tgt_ps = self.get_path_spec(url=self.url)

        stat = self._adaptor.stat(self.shell, tgt_ps)
        if stat['mode'] == 'dir':
            return True
        else:
            return False

# ------------------------------------------------------------------------------





__author__    = "Andre Merzky, Ole Weidner, Thomas Schatz"
__copyright__ = "Copyright 2012-2013, The SAGA Project"
__license__   = "MIT"


""" SAGA job description interface """

import radical.utils.signatures   as rus

from ..         import attributes as sa
from .constants import *


# TODO: file transfer to accept strings *and* dicts like in RP
# TODO: flags: cacheable: don't stage twice in same session
#              remove   : remove src after staging


# ------------------------------------------------------------------------------
#
class Description (sa.Attributes) :
    """ The job description class. """

    # --------------------------------------------------------------------------
    #
    @rus.takes   ('Description')
    @rus.returns (rus.nothing)
    def __init__(self):

        # set attribute interface properties

        self._attributes_extensible  (True)
        self._attributes_camelcasing (True)

        # register properties with the attribute interface

        self._attributes_register  (EXECUTABLE           , None, sa.STRING, sa.SCALAR, sa.WRITEABLE)
        self._attributes_register  (PRE_EXEC             , None, sa.STRING, sa.VECTOR, sa.WRITEABLE)
        self._attributes_register  (POST_EXEC            , None, sa.STRING, sa.VECTOR, sa.WRITEABLE)
        self._attributes_register  (ARGUMENTS            , None, sa.STRING, sa.VECTOR, sa.WRITEABLE)
        self._attributes_register  (ENVIRONMENT          , None, sa.STRING, sa.DICT,   sa.WRITEABLE)
        self._attributes_register  (TOTAL_CPU_COUNT      , None, sa.INT,    sa.SCALAR, sa.WRITEABLE)
        self._attributes_register  (TOTAL_GPU_COUNT      , None, sa.INT,    sa.SCALAR, sa.WRITEABLE)
        self._attributes_register  (NUMBER_OF_PROCESSES  , None, sa.INT,    sa.SCALAR, sa.WRITEABLE)
        self._attributes_register  (PROCESSES_PER_HOST   , None, sa.INT,    sa.SCALAR, sa.WRITEABLE)
        self._attributes_register  (THREADS_PER_PROCESS  , None, sa.INT,    sa.SCALAR, sa.WRITEABLE)
        self._attributes_register  (WORKING_DIRECTORY    , None, sa.STRING, sa.SCALAR, sa.WRITEABLE)
        self._attributes_register  (INTERACTIVE          , None, sa.BOOL,   sa.SCALAR, sa.WRITEABLE)
        self._attributes_register  (INPUT                , None, sa.STRING, sa.SCALAR, sa.WRITEABLE)
        self._attributes_register  (OUTPUT               , None, sa.STRING, sa.SCALAR, sa.WRITEABLE)
        self._attributes_register  (ERROR                , None, sa.STRING, sa.SCALAR, sa.WRITEABLE)
        self._attributes_register  (FILE_TRANSFER        , None, sa.ANY,    sa.VECTOR, sa.WRITEABLE)
        self._attributes_register  (CLEANUP              , None, sa.BOOL,   sa.SCALAR, sa.WRITEABLE)
        self._attributes_register  (JOB_START_TIME       , None, sa.TIME,   sa.SCALAR, sa.WRITEABLE)
        self._attributes_register  (WALL_TIME_LIMIT      , None, sa.INT,    sa.SCALAR, sa.WRITEABLE)
        self._attributes_register  (TOTAL_PHYSICAL_MEMORY, None, sa.INT,    sa.SCALAR, sa.WRITEABLE)
        self._attributes_register  (SYSTEM_ARCHITECTURE  ,   {}, sa.STRING, sa.DICT,   sa.WRITEABLE)
        self._attributes_register  (OPERATING_SYSTEM_TYPE, None, sa.ENUM,   sa.SCALAR, sa.WRITEABLE)
        self._attributes_register  (CANDIDATE_HOSTS      , None, sa.STRING, sa.VECTOR, sa.WRITEABLE)
        self._attributes_register  (QUEUE                , None, sa.STRING, sa.SCALAR, sa.WRITEABLE)
        self._attributes_register  (NAME                 , None, sa.STRING, sa.SCALAR, sa.WRITEABLE)
        self._attributes_register  (PROJECT              , None, sa.STRING, sa.SCALAR, sa.WRITEABLE)
        self._attributes_register  (JOB_CONTACT          , None, sa.STRING, sa.VECTOR, sa.WRITEABLE)
        self._attributes_register  (SPMD_VARIATION       , None, sa.ENUM,   sa.SCALAR, sa.WRITEABLE)

        self._attributes_set_enums (SPMD_VARIATION,      ['MPI', 'OpenMP', 'MPICH-G'])

        self._env_is_list = False

        self._attributes_set_getter (ENVIRONMENT, self._get_env)
        self._attributes_set_setter (ENVIRONMENT, self._set_env)


    # --------------------------------------------------------------------------
    #
    def _set_env (self, val) :
        if  isinstance (val, list) :
            self._env_is_list = True


    # --------------------------------------------------------------------------
    #
    def _get_env (self) :
        env = self.get_attribute (ENVIRONMENT)
        if  self._env_is_list :
            self._env_is_list = False
            return ["%s=%s" % (key, val) for (key, val) in list(env.items ())]
        return env


    # --------------------------------------------------------------------------
    #
    @rus.takes   ('Description',
                  ('Description', dict))
    @rus.returns ('Description')
    def __deepcopy__ (self, memo) :
        other = Description ()
        return self.clone (other)


    # --------------------------------------------------------------------------
    #
    @rus.takes   ('Description',
                  'Description')
    @rus.returns ('Description')
    def clone (self, other=None) :
        """
        clone()

        Implements deep copy. u

        Unlike the default python assignment (copy object reference),
        a deep copy will create a new object instance with the same state --
        after a deep copy, a change on one instance will not affect the other.
        """

        # a job description only has attributes - so create a new instance,
        # clone the attribs, and done.
        if not other :
            other = Description ()

        return self._attributes_deep_copy (other)


# ------------------------------------------------------------------------------





__author__    = "Andre Merzky"
__copyright__ = "Copyright 2012-2013, The SAGA Project"
__license__   = "MIT"


from ...           import exceptions as rse

from .decorators   import CPI_SYNC_CALL  as SYNC


class Attributes (object) :

    @SYNC
    def attribute_getter    (self, key)         : pass

    @SYNC
    def attribute_setter    (self, key, val)    : pass

    @SYNC
    def attribute_lister    (self)              : pass

    @SYNC
    def attribute_caller    (self, key, id, cb) : pass

    @SYNC
    def add_callback        (self, key, cb)     :
        raise rse.NotImplemented ("Callbacks not supported for this backend")









__author__    = "Andre Merzky"
__copyright__ = "Copyright 2012-2013, The SAGA Project"
__license__   = "MIT"


""" Provides a number of call decorators. """

import re
import inspect

from ...task       import Task  as T_Task   # class
from ...constants  import SYNC  as T_SYNC
from ...constants  import ASYNC as T_ASYNC
from ...constants  import TASK  as T_TASK   # flag

from ...           import exceptions as rse


# ------------------------------------------------------------------------------
# decorator, which switches method to
# _async version if ttype is set and !None
def SYNC_CALL (sync_function) :

    def wrap_function (self, *args, **kwargs) :

        if 'ttype' in kwargs and kwargs['ttype']:

            if not kwargs['ttype'] in (T_SYNC, T_ASYNC, T_TASK) :
                # cannot handle that ttype value, do not call async methods
                ttype = kwargs['ttype']
                msg   = " %s: async %s() called with invalid tasktype (%s)" \
                      % (self.__class__.__name__, sync_function.__name__,
                         str(ttype))
                raise rse.BadParameter (msg)

            # call async method flavor
            try :
                async_function_name = "%s_async"  %  sync_function.__name__
                async_function      = getattr (self, async_function_name)

            except AttributeError as e:
                msg = " %s: async %s() not implemented" \
                    % (self.__class__.__name__, sync_function.__name__)
                raise rse.NotImplemented (msg) from e

            else :
                # 'self' not needed, getattr() returns member function
                return async_function (*args, **kwargs)

        # no ttype, or ttype==None -- make sure it's gone, and call default sync
        # function
        if 'ttype' in kwargs :
            del kwargs['ttype']

        # only some functions will provide metrics, and thus need the _from_task
        # parameter -- strip that as well if its not needed
        if '_from_task' in kwargs:
            if '_from_task' not in inspect.getargspec (sync_function).args:
                del(kwargs['_from_task'])

        return sync_function (self, *args, **kwargs)

    return wrap_function


# ------------------------------------------------------------------------------
# we assume that async calls want to call, aehm, async methods...
def ASYNC_CALL (async_function) :
    return async_function


# ------------------------------------------------------------------------------
# sync cpi calls are only called when an adaptor does not implement that call --
# we thus raise a NotImplemented exception.
#
def CPI_SYNC_CALL (cpi_sync_function) :

    def wrap_function (self, *args, **kwargs) :
        raise rse.NotImplemented ("%s.%s is not implemented for %s.%s (%s)"
                %  (self.get_api ().__class__.__name__,
                    inspect.stack ()[1][3],
                    self._adaptor._name,
                    self.__class__.__name__,
                    inspect.getmembers (cpi_sync_function)[15][1]))

    return wrap_function


# ------------------------------------
# async cpi calls attempt to wrap sync adaptor calls into threaded tasks
def CPI_ASYNC_CALL (cpi_async_function) :

    def wrap_function (self, *args, **kwargs) :


        if 'ttype' not in kwargs :
            msg = " %s: async %s() called with no tasktype" \
                % (self.__class__.__name__, cpi_async_function.__name__)
            raise rse.BadParameter (msg)


        ttype = kwargs['ttype']
        del kwargs['ttype']


        if ttype not in (T_SYNC, T_ASYNC, T_TASK) :
            # cannot handle that ttype value, do not call async methods
            msg = " %s: async %s() called with invalid tasktype (%s)" \
                % (self.__class__.__name__, cpi_async_function.__name__,
                   str(ttype))
            raise rse.BadParameter (msg)


        cpi_sync_function_name = None
        cpi_sync_function      = None

        # find sync method flavor
        try :
            cpi_sync_function_name = re.sub ("_async$", "",
                                             cpi_async_function.__name__)
            cpi_sync_function      = getattr (self, cpi_sync_function_name)

        except AttributeError as e:
            msg = " %s: sync %s() not implemented" \
                % (self.__class__.__name__, cpi_sync_function.__name__)
            raise rse.NotImplemented (msg) from e


        # got the sync call, wrap it in a task
        c = {'_call'   : cpi_sync_function,
             '_args'   : args,
             '_kwargs' : kwargs }   # no ttype!

        return T_Task(self, cpi_sync_function_name, c, ttype)

    return wrap_function


# ------------------------------------------------------------------------------





__author__    = "Andre Merzky, Ole Weidner"
__copyright__ = "Copyright 2012-2013, The SAGA Project"
__license__   = "MIT"


""" SAGA job service interface """


import radical.utils            as ru
import radical.utils.signatures as rus

from ..constants import SYNC, ASYNC, TASK
from ..adaptors  import base    as sab

from .. import task             as st
from .. import base             as sb
from .. import sasync
from .. import exceptions       as se
from .. import session          as ss

from .  import job              as j
from .  import description      as descr


# ------------------------------------------------------------------------------
#
class Service (sb.Base, sasync.Async) :
    """
    The job.Service represents a resource management backend, and as such allows
    the creation, submission and management of jobs.

    A job.Service represents anything which accepts job creation requests, and
    which manages thus created :class:`saga.job.Job` instances.  That can be
    a local shell, a remote ssh shell, a cluster queuing system, a IaaS backend
    -- you name it.

    The job.Service is identified by an URL, which usually points to the contact
    endpoint for that service.


    Example::

        service  = saga.job.Service("fork://localhost")
        ids = service.list()

        for job_id in ids :
            print(job_id)

            j = service.get_job(job_id)

            if j.get_state() == saga.job.Job.Pending:
                print("pending")
            elif j.get_state() == saga.job.Job.Running:
                print("running")
            else:
                print("job is already final!")

        service.close()
    """

    # --------------------------------------------------------------------------
    #
    @rus.takes   ('Service',
                  rus.optional ((str, ru.Url)),
                  rus.optional (ss.Session),
                  rus.optional (sab.Base),
                  rus.optional (dict),
                  rus.optional (rus.one_of (SYNC, ASYNC, TASK)))
    @rus.returns (rus.nothing)
    def __init__ (self, rm=None, session=None,
                  _adaptor=None, _adaptor_state={}, _ttype=None) :
        """
        __init__(rm, session)

        Create a new job.Service instance.

        :param rm:      resource manager URL
        :type  rm:      string or :class:`saga.Url`
        :param session: an optional session object with security contexts
        :type  session: :class:`saga.Session`
        :rtype:         :class:`saga.job.Service`
        """

        # job service instances are resource hogs.  Before attempting to create
        # a new instance, we attempt to clear out all old instances.   There is
        # some collateral damage: we cannot run the Python GC over only the
        # job.Service instances, but have to run it globally -- however,
        # compared to the latency introduced by the job service setup, this
        # should be a minor inconvenienve (tm)
        try :
            import gc
            gc.collect ()

        except :
            pass


        # param checks
        self.valid  = False
        url         = ru.Url (rm)

        if  not url.scheme :
            url.scheme = 'fork'

        if  not url.host :
            url.host = 'localhost'

        if  not session :
            session = ss.Session (default=True)

        scheme = url.scheme.lower ()

        self._super = super  (Service, self)
        self._super.__init__ (scheme, _adaptor, _adaptor_state,
                              url, session, ttype=_ttype)

        self.valid  = True

    # --------------------------------------------------------------------------
    #
    @classmethod
    @rus.takes   ('Service',
                  rus.optional ((ru.Url, str)),
                  rus.optional (ss.Session),
                  rus.optional (rus.one_of (SYNC, ASYNC, TASK)))
    @rus.returns (st.Task)
    def create   (cls, rm=None, session=None, ttype=SYNC) :
        """
        create(rm=None, session=None)
        Create a new job.Service instance asynchronously.

        :param rm:      resource manager URL
        :type  rm:      string or :class:`saga.Url`
        :param session: an optional session object with security contexts
        :type  session: :class:`saga.Session`
        :rtype:         :class:`saga.Task`
        """

        # param checks
        if not session :
            session = ss.Session (default=True)

        url = ru.Url (rm)

        return cls (url, session, _ttype=ttype)._init_task


    # --------------------------------------------------------------------------
    #
    @rus.takes     ('Service')
    @rus.returns   (str)
    def __str__ (self):
        """
        __str__()

        String representation. Returns the job service Url.
        """

        if  self.valid :
            return "[%s]" % self.url

        return ""


    # --------------------------------------------------------------------------
    #
    @rus.takes     ('Service')
    @rus.returns   (rus.nothing)
    def close (self) :
        """
        close()

        Close the job service instance and disconnect from the (remote)
        job service if necessary. Any subsequent calls to a job service
        instance after `close()` was called will fail.

        Example::

            service = saga.job.Service("fork://localhost")

            # do something with the 'service' object, create jobs, etc...

            service.close()

            service.list() # this call will throw an exception


        .. warning:: While in principle the job service destructor calls
            `close()` automatically when a job service instance goes out of
            scope, you **shouldn't rely on it**. Python's garbage collection can
            be a bit odd at times, so you should always call `close()`
            explicitly.  Especially in a **multi-threaded program** this will
            help to avoid random errors.  """

        if not self.valid :
            raise se.IncorrectState ("This instance was already closed.")

        self._adaptor.close ()
        self.valid = False


    # --------------------------------------------------------------------------
    #
    @rus.takes     ('Service',
                    descr.Description,
                    rus.optional (rus.one_of (SYNC, ASYNC, TASK)))
    @rus.returns   ((j.Job, st.Task))
    def create_job (self, job_desc, ttype=None) :
        """
        create_job(job_desc)

        Create a new job.Job instance from a :class:`~saga.job.Description`. The
        resulting job instance is in :data:`~saga.job.NEW` state.

        :param job_desc: job description to create the job from
        :type job_desc:  :data:`saga.job.Description`
        :param ttype: |param_ttype|
        :rtype:       :class:`saga.job.Job` or |rtype_ttype|

        create_job() accepts a job description, which described the
        application instance to be created by the backend.  The create_job()
        method is not actually attempting to *run* the job, but merely parses
        the job description for syntactic and semantic consistency.  The job
        returned object is thus not in 'Pending' or 'Running', but rather in
        'New' state.  The actual submission is performed by calling run() on
        the job object.


        Example::

            # A job.Description object describes the executable/application
            # and its requirements
            job_desc = saga.job.Description()
            job_desc.executable  = '/bin/sleep'
            job_desc.arguments   = ['10']
            job_desc.output      = 'myjob.out'
            job_desc.error       = 'myjob.err'

            service = saga.job.Service('local://localhost')

            job = service.create_job(job_desc)

            # Run the job and wait for it to finish
            job.run()
            print("Job ID    : %s" % (job.job_id))
            job.wait()

            # Get some info about the job
            print("Job State : %s" % (job.state))
            print("Exitcode  : %s" % (job.exit_code))

            service.close()
        """


        if not self.valid :
            raise se.IncorrectState ("This instance was already closed.")

        jd_copy = descr.Description()
        job_desc._attributes_deep_copy (jd_copy)

        # do some sanity checks: if the adaptor has specified a set of supported
        # job description attributes, we scan the given description for any
        # mismatches, and complain then.
        adaptor_info = self._adaptor._adaptor.get_info ()

        if  'capabilities'    in adaptor_info             and \
            'jdes_attributes' in adaptor_info['capabilities'] :

            # this is the list of key supported by the adaptor.  These
            # attributes may be set to non-default values
            supported_keys = adaptor_info['capabilities']['jdes_attributes']

            # use an empty job description to compare default values
            jd_default = descr.Description ()

            for key in jd_copy.list_attributes () :

                val     = jd_copy   .get_attribute (key)
                default = jd_default.get_attribute (key)

                # Also, we make string compares case insensitive
                if isinstance (val,     str) : val     = val    .lower ()
                if isinstance (default, str) : default = default.lower ()

                # supported keys are also valid, as are keys with default or
                # None values
                if  key not in supported_keys and \
                    val != default            and \
                    val                       :

                    msg = "'JobDescription.%s' (%s) not supported by %s" \
                        % (key, val, adaptor_info['name'])
                    raise se.BadParameter._log (self._logger, msg)


        # make sure at least 'executable' is defined
        if jd_copy.executable is None:
            raise se.BadParameter("No executable defined")

        # convert environment to string
        if jd_copy.attribute_exists ('Environment') :
            for (key, value) in jd_copy.environment.items():
                jd_copy.environment[key] = str(value)

        return self._adaptor.create_job (jd_copy, ttype=ttype)


    # --------------------------------------------------------------------------
    #
    @rus.takes   ('Service',
                  str,
                  rus.optional (str),
                  rus.optional (rus.one_of (SYNC, ASYNC, TASK)))
    @rus.returns ((j.Job, st.Task))
    def run_job  (self, cmd, host=None, ttype=None) :
        """
        run_job(cmd, host=None)
        """

        if not self.valid :
            raise se.IncorrectState ("This instance was already closed.")

        if not cmd:
            raise se.BadParameter('run_job needs a command to run.  Duh!')

        try:
            # lets see if the adaptor implements run_job
            return self._adaptor.run_job (cmd, host, ttype=ttype)
        except:
            # fall back to the default implementation below
            pass

        # The adaptor has no run_job -- we here provide a generic implementation
        # FIXME: split should be more clever and respect POSIX shell syntax.
        args = cmd.split()

        jd = descr.Description()
        jd.executable = args[0]
        jd.arguments  = args[1:]

        job = self.create_job(jd)
        job.run()

        return job


    # --------------------------------------------------------------------------
    #
    @rus.takes   ('Service',
                  rus.optional (rus.one_of (SYNC, ASYNC, TASK)))
    @rus.returns ((rus.list_of (str), st.Task))
    def list     (self, ttype=None) :
        """
        list()

        Return a list of the jobs that are managed by this Service
        instance.

        .. seealso::
           The :data:`~saga.job.Service.jobs` property and the
           :meth:`~saga.job.Service.list` method are semantically
           equivalent.

        :ttype: |param_ttype|
        :rtype: list of :class:`saga.job.Job`

        As the job.Service represents a job management backend, list() will
        return a list of job IDs for all jobs which are known to the backend,
        and which can potentially be accessed and managed by the application.


        Example::

            service  = saga.job.Service("fork://localhost")
            ids = service.list()

            for job_id in ids :
                print(job_id)

            service.close()
        """

        if not self.valid :
            raise se.IncorrectState ("This instance was already closed.")


        return self._adaptor.list (ttype=ttype)

    jobs = property (list)


    # --------------------------------------------------------------------------
    #
    @rus.takes   ('Service',
                  rus.optional (rus.one_of (SYNC, ASYNC, TASK)))
    @rus.returns ((ru.Url, st.Task))
    def get_url  (self, ttype=None) :
        """
        get_url()

        Return the URL this Service instance was created with.

        .. seealso::
           The :data:`~saga.job.Service.url` property and the
           :meth:`~saga.job.Service.get_url` method are semantically
           equivalent and only duplicated for convenience.
        """

        if not self.valid :
            raise se.IncorrectState ("This instance was already closed.")

        return self._adaptor.get_url (ttype=ttype)

    url = property (get_url)


    # --------------------------------------------------------------------------
    #
    @rus.takes   ('Service',
                  str,
                  rus.optional (rus.one_of (SYNC, ASYNC, TASK)))
    @rus.returns ((j.Job, st.Task))
    def get_job  (self, job_id, ttype=None) :
        """
        get_job(job_id)

        Return the job object for a given job id.

        :param job_id: The id of the job to retrieve
        :rtype:        :class:`saga.job.Job`


        Job objects are a local representation of a remote stateful entity.
        The job.Service supports to reconnect to those remote entities::

            service = saga.job.Service("fork://localhost")
            j  = service.get_job(my_job_id)

            if j.get_state() == saga.job.Job.Pending:
                print("pending")
            elif j.get_state() == saga.job.Job.Running:
                print("running")
            else:
                print("job is already final!")

            service.close()
        """

        if not self.valid :
            raise se.IncorrectState ("This instance was already closed.")

        return self._adaptor.get_job (job_id, ttype=ttype)

# FIXME: add get_self()

# ------------------------------------------------------------------------------





__author__    = "Ole Weidner"
__copyright__ = "Copyright 2012-2013, The SAGA Project"
__license__   = "MIT"


import os
import sys
import time

from PIL import Image

import radical.saga as rs


#-----------------------------------------------------------------------------
#

# Change REMOTE_HOST to the machine you want to run this on.
# You might have to change the URL scheme below for REMOTE_JOB_ENDPOINT
# accordingly.
REMOTE_HOST = "localhost"  # try this with different hosts

# This refers to your working directory on 'REMOTE_HOST'. If you use a\
# cluster for 'REMOTE_HOST', make sure this points to a shared filesystem.
REMOTE_DIR = "/tmp/"  # change this to your home directory

# If you change 'REMOTE_HOST' above, you might have to change 'ssh://' to e.g.,
# 'pbs+ssh://', 'sge+ssh://', depdending on the type of service endpoint on
# that particualr host.
REMOTE_JOB_ENDPOINT = "ssh://" + REMOTE_HOST

# At the moment radical.saga only provides an sftp file adaptor, so changing
# the URL scheme here wouldn't make any sense.
REMOTE_FILE_ENDPOINT = "sftp://" + REMOTE_HOST + "/" + REMOTE_DIR

# the dimension (in pixel) of the whole fractal
imgx = 2048
imgy = 2048

# the number of tiles in X and Y direction
tilesx = 2
tilesy = 2

#-----------------------------------------------------------------------------
#
if __name__ == "__main__":
    try:
        # Your ssh identity on the remote machine
        ctx = rs.Context("ssh")
        #ctx.user_id = ""

        session = rs.Session()
        session.add_context(ctx)

        # list that holds the jobs
        jobs = []

        # create a working directory in /scratch
        dirname = '%s/mbrot/' % (REMOTE_FILE_ENDPOINT)
        workdir = rs.filesystem.Directory(dirname, rs.filesystem.CREATE,
                                            session=session)

        # copy the executable and warpper script to the remote host
        mbwrapper = rs.filesystem.File('file://localhost/%s/mandelbrot.sh' % os.getcwd())
        mbwrapper.copy(workdir.get_url())
        mbexe = rs.filesystem.File('file://localhost/%s/mandelbrot.py' % os.getcwd())
        mbexe.copy(workdir.get_url())

        # the saga job services connects to and provides a handle
        # to a remote machine. In this case, it's your machine.
        # fork can be replaced with ssh here:
        jobservice = rs.job.Service(REMOTE_JOB_ENDPOINT, session=session)

        for x in range(0, tilesx):
            for y in range(0, tilesy):

                # describe a single Mandelbrot job. we're using the
                # directory created above as the job's working directory
                outputfile = 'tile_x%s_y%s.gif' % (x, y)
                jd = rs.job.Description()
                #jd.queue             = "development"
                jd.wall_time_limit   = 10
                jd.total_cpu_count   = 1
                jd.working_directory = workdir.get_url().path
                jd.executable        = 'sh'
                jd.arguments         = ['mandelbrot.sh', imgx, imgy,
                                        int(imgx/tilesx*x), int(imgx/tilesx*(x+1)),
                                        int(imgy/tilesy*y), int(imgy/tilesy*(y+1)),
                                        outputfile]
                # create the job from the description
                # above, launch it and add it to the list of jobs
                job = jobservice.create_job(jd)
                job.run()
                jobs.append(job)
                print(' * Submitted %s. Output will be written to: %s' % (job.id, outputfile))

        # wait for all jobs to finish
        while len(jobs) > 0:
            for job in jobs:
                jobstate = job.get_state()
                print(' * Job %s status: %s' % (job.id, jobstate))
                if jobstate in [rs.job.DONE, rs.job.FAILED]:
                    jobs.remove(job)
            print("")
            time.sleep(5)

        # copy image tiles back to our 'local' directory
        for image in workdir.list('*.gif'):
            print(' * Copying %s/%s/%s back to %s' % (REMOTE_FILE_ENDPOINT,
                                                      workdir.get_url(),
                                                      image, os.getcwd()))
            workdir.copy(image, 'file://localhost/%s/' % os.getcwd())

        # stitch together the final image
        fullimage = Image.new('RGB', (imgx, imgy), (255, 255, 255))
        print(' * Stitching together the whole fractal: mandelbrot_full.gif')
        for x in range(0, tilesx):
            for y in range(0, tilesy):
                partimage = Image.open('tile_x%s_y%s.gif' % (x, y))
                fullimage.paste(partimage,
                                (int(imgx/tilesx*x), int(imgy/tilesy*y),
                                 int(imgx/tilesx*(x+1)), int(imgy/tilesy*(y+1))))
        fullimage.save("mandelbrot_full.gif", "GIF")
        sys.exit(0)

    except rs.SagaException as ex:
        # Catch all saga exceptions
        print("An exception occured: (%s) %s " % (ex.type, (str(ex))))
        # Trace back the exception. That can be helpful for debugging.
        print(" \n*** Backtrace:\n %s" % ex.traceback)
        sys.exit(-1)

    except KeyboardInterrupt:
    # ctrl-c caught: try to cancel our jobs before we exit
        # the program, otherwise we'll end up with lingering jobs.
        for job in jobs:
            job.cancel()
        sys.exit(-1)





__author__    = "Andre Merzky, Ole Weidner"
__copyright__ = "Copyright 2012-2013, The SAGA Project"
__license__   = "MIT"


import radical.utils              as ru
import radical.utils.signatures   as rus

from ..constants import SYNC, ASYNC, TASK
from ..adaptors  import base      as sab
from ..namespace import directory as nsdir

from .. import attributes         as sa
from .. import session            as ss
from .. import task               as st
from .  import constants          as c


# ------------------------------------------------------------------------------
#
class LogicalDirectory (nsdir.Directory, sa.Attributes) :

    # --------------------------------------------------------------------------
    #
    @rus.takes   ('LogicalDirectory',
                  rus.optional ((ru.Url, str)),
                  rus.optional (int, rus.nothing),
                  rus.optional (ss.Session),
                  rus.optional (sab.Base),
                  rus.optional (dict),
                  rus.optional (rus.one_of (SYNC, ASYNC, TASK)))
    @rus.returns (rus.nothing)
    def __init__ (self, url=None, flags=c.READ, session=None,
                  _adaptor=None, _adaptor_state={}, _ttype=None) :
        '''
        __init__(url, flags=READ, session=None)

        Create a new Logical Directory instance.

        url:       saga.Url
        flags:     flags enum
        session:   saga.Session
        ret:       obj
        '''

        # param checks
        if not flags : flags = 0
        url = ru.Url (url)

        self._nsdirec = super  (LogicalDirectory, self)
        self._nsdirec.__init__ (url, flags, session,
                                _adaptor, _adaptor_state, _ttype=_ttype)


    # --------------------------------------------------------------------------
    #
    @classmethod
    @rus.takes   ('LogicalDirectory',
                  rus.one_of (ru.Url, str),
                  rus.optional (int, rus.nothing),
                  rus.optional (ss.Session),
                  rus.optional (rus.one_of (SYNC, ASYNC, TASK)))
    @rus.returns (st.Task)
    def create (cls, url, flags=c.READ, session=None, ttype=None) :
        '''
        url:       saga.Url
        flags:     saga.replica.flags enum
        session:   saga.Session
        ttype:     saga.task.type enum
        ret:       saga.Task
        '''

        if not flags : flags = 0
        _nsdirec = super (LogicalDirectory, cls)
        return _nsdirec.create (url, flags, session, ttype=ttype)


    @rus.takes   ('LogicalDirectory',
                  rus.one_of (ru.Url, str),
                  rus.optional (rus.one_of (SYNC, ASYNC, TASK)))
    @rus.returns ((bool, st.Task))
    def is_file (self, tgt=None, ttype=None) :
        '''
        is_file(tgt=None)

        tgt:           saga.Url / string
        ttype:          saga.task.type enum
        ret:            bool / saga.Task
        '''
        if tgt: return self._adaptor.is_file      (tgt, ttype=ttype)
        else  : return self._nsdirec.is_entry_self(     ttype=ttype)


    @rus.takes   ('LogicalDirectory',
                  rus.one_of (ru.Url, str),
                  rus.optional (int, rus.nothing),
                  rus.optional (rus.one_of (SYNC, ASYNC, TASK)))
    @rus.returns (('LogicalFile', st.Task))
    def open (self, tgt, flags=c.READ, ttype=None) :
        '''
        open(tgt, flags=READ)

        tgt:      saga.Url
        flags:    saga.namespace.flags enum
        ttype:    saga.task.type enum
        ret:      saga.namespace.Entry / saga.Task
        '''
        if not flags : flags = 0
        tgt_url = ru.Url (tgt)
        return self._adaptor.open (tgt_url, flags, ttype=ttype)


    # ----------------------------------------------------------------
    #
    @rus.takes   ('LogicalDirectory',
                  rus.one_of (ru.Url, str),
                  rus.optional (int, rus.nothing),
                  rus.optional (rus.one_of (SYNC, ASYNC, TASK)))
    @rus.returns (('LogicalDirectory', st.Task))
    def open_dir (self, tgt, flags=c.READ, ttype=None) :
        '''
        open_dir(tgt, flags=READ)

        :param tgt:   name/path of the directory to open
        :param flags: directory creation flags

        ttype:    saga.task.type enum
        ret:      saga.namespace.Directory / saga.Task

        Open and return a new directoy

           The call opens and returns a directory at the given location.

           Example::

               # create a subdir 'data' in /tmp
               dir = saga.namespace.Directory("sftp://localhost/tmp/")
               data = dir.open_dir ('data/', saga.namespace.Create)
        '''
        if not flags : flags = 0
        tgt_url = ru.Url (tgt)
        return self._adaptor.open_dir (tgt_url, flags, ttype=ttype)


    # ----------------------------------------------------------------
    #
    # replica methods
    #
    # --------------------------------------------------------------------------
    #
    @rus.takes   ('LogicalDirectory',
                  rus.one_of (ru.Url, str),
                  rus.optional (rus.one_of (SYNC, ASYNC, TASK)))
    @rus.returns ((int, st.Task))
    def get_size (self, tgt, ttype=None) :
        '''
        get_size(tgt)

        tgt:     logical file to get size for
        ttype:    saga.task.type enum
        ret:      int / saga.Task

        Returns the size of the physical file represented by the given logical
        file (in bytes)

           Example::

               # get a logical directory handle
               lf = saga.replica.LogicalFile("irods://localhost/tmp/data/")

               # print a logical file's size
               print(lf.get_size ('data.dat'))

        '''
        tgt_url = ru.Url (tgt)
        return self._adaptor.get_size (tgt_url, ttype=ttype)


    # ----------------------------------------------------------------
    #
    @rus.takes   ('LogicalDirectory',
                  rus.optional (str),
                  rus.optional (str),
                  rus.optional (int, rus.nothing),
                  rus.optional (rus.one_of (SYNC, ASYNC, TASK)))
    @rus.returns ((rus.list_of (ru.Url), st.Task))
    def find (self, name_pattern, attr_pattern=None, flags=c.RECURSIVE,
              ttype=None) :
        '''
        find(name_pattern, attr_pattern=None, flags=RECURSIVE)

        name_pattern:   string
        attr_pattern:   string
        flags:          flags enum
        ttype:          saga.task.type enum
        ret:            list [saga.Url] / saga.Task

        '''
        if not flags: flags = 0

        if attr_pattern:
            return self._adaptor.find_replicas(name_pattern, attr_pattern,
                                               flags, ttype=ttype)
        else:
            return self._nsdirec.find(name_pattern, flags, ttype=ttype)


# ------------------------------------------------------------------------------





__author__    = "Andre Merzky"
__copyright__ = "Copyright 2012-2013, The SAGA Project"
__license__   = "MIT"


import radical.saga.adaptors.cpi.decorators as cpi_dec
import radical.saga.adaptors.cpi.namespace  as cpi_ns
import radical.saga.adaptors.cpi.attributes as cpi_att

SYNC  = cpi_dec.CPI_SYNC_CALL
ASYNC = cpi_dec.CPI_ASYNC_CALL


# keep order of inheritance!  super() below uses MRO
class LogicalFile (cpi_ns.entry.Entry, 
                   cpi_att.Attributes) :

    # ----------------------------------------------------------------
    #
    # initialization methods
    #
    def __init__ (self, api, adaptor) :

        self._cpi_nsentry = super  (LogicalFile, self)
        self._cpi_nsentry.__init__ (api, adaptor)


    @SYNC
    def init_instance         (self, url, flags, session)      : pass
    @ASYNC
    def init_instance_async   (self, url, flags, session)      : pass


    # ----------------------------------------------------------------
    #
    # replica methods
    #
    @SYNC
    def is_file_self          (self, ttype)                    : pass
    @ASYNC
    def is_file_self_async    (self, ttype)                    : pass

    @SYNC
    def get_size_self         (self, ttype)                    : pass
    @ASYNC
    def get_size_self_async   (self, ttype)                    : pass

    @SYNC
    def add_location          (self, name, ttype)              : pass
    @ASYNC
    def add_location_async    (self, name, ttype)              : pass

    @SYNC
    def remove_location       (self, name, ttype)              : pass
    @ASYNC
    def remove_location_async (self, name, ttype)              : pass

    @SYNC
    def update_location       (self, old, new, ttype)          : pass
    @ASYNC
    def update_location_async (self, old, new, ttype)          : pass

    @SYNC
    def list_locations        (self, ttype)                    : pass
    @ASYNC
    def list_locations_async  (self, ttype)                    : pass

    @SYNC
    def replicate             (self, name, flags, ttype)       : pass
    @ASYNC
    def replicate_async       (self, name, flags, ttype)       : pass

    @SYNC
    def upload                (self, name, tgt, flags, ttype)  : pass
    @ASYNC
    def upload_async          (self, name, tgt, flags, ttype)  : pass

    @SYNC
    def download              (self, name, src, flags, ttype)  : pass
    @ASYNC
    def download_async        (self, name, src, flags, ttype)  : pass








__author__    = "Andre Merzky, Ole Weidner"
__copyright__ = "Copyright 2012-2013, The SAGA Project"
__license__   = "MIT"


""" file adaptor implementation on top tof the HTTP protocol
"""

import os
import urllib.request, urllib.parse, urllib.error

import radical.utils as ru

from ...exceptions           import *
from ...                     import url        as rsurl
from ...utils                import pty_shell  as rsups
from ...utils                import misc       as rsumisc
from ...                     import filesystem as api_fs
from ...filesystem.constants import *
from ..                      import base       as a_base
from ..cpi                   import filesystem as cpi_fs
from ..cpi                   import decorators as cpi_decs

SYNC_CALL  = cpi_decs.SYNC_CALL
ASYNC_CALL = cpi_decs.ASYNC_CALL


# --------------------------------------------------------------------
# the adaptor name
#
_ADAPTOR_NAME          = "radical.saga.adaptors.http_file"
_ADAPTOR_SCHEMAS       = ["http", "https"]
_ADAPTOR_OPTIONS       = []

# --------------------------------------------------------------------
# the adaptor capabilities & supported attributes
#
_ADAPTOR_CAPABILITIES  = {
    "metrics"          : [],
    "contexts"         : {"userpass" : "username/password pair for ssh"}
}

# --------------------------------------------------------------------
# the adaptor documentation
#
_ADAPTOR_DOC           = {
    "name"             : _ADAPTOR_NAME,
    "cfg_options"      : _ADAPTOR_OPTIONS,
    "capabilities"     : _ADAPTOR_CAPABILITIES,
    "description"      : """The HTTP file adpator allows file transfer (copy) from remote resources to the local machine via the HTTP/HTTPS protocol, similar to cURL.""",
    "example"          : "examples/files/http_file_copy.py",
    "schemas"          : {"http"   :"use the http protocol to access a remote file",
                          "https"  :"use the https protocol to access a remote file"}
}

# --------------------------------------------------------------------
# the adaptor info is used to register the adaptor with SAGA

_ADAPTOR_INFO = {
    "name": _ADAPTOR_NAME,
    "version": "v0.1",
    "schemas": _ADAPTOR_SCHEMAS,
    "cpis": [
        {
            "type": "radical.saga.namespace.Entry",
            "class": "HTTPFile"
        },
        {
            "type": "radical.saga.filesystem.File",
            "class": "HTTPFile"
        }
    ]
}


###############################################################################
# The adaptor class
class Adaptor (a_base.Base):
    """
    This is the actual adaptor class, which gets loaded by SAGA (i.e. by the
    SAGA engine), and which registers the CPI implementation classes which
    provide the adaptor's functionality.
    """

    # ----------------------------------------------------------------
    #
    def __init__(self):

        a_base.Base.__init__(self, _ADAPTOR_INFO, _ADAPTOR_OPTIONS)

    # ----------------------------------------------------------------
    #
    def sanity_check(self):
        pass


###############################################################################
#
class HTTPFile (cpi_fs.File):
    """ Implements radical.saga.adaptors.cpi.filesystem.File
    """
    # ----------------------------------------------------------------
    #
    def __init__(self, api, adaptor):

        self._cpi_base = super(HTTPFile, self)
        self._cpi_base.__init__(api, adaptor)

    # ----------------------------------------------------------------
    #
    def __del__(self):
        self.finalize(kill=True)

    # ----------------------------------------------------------------
    #
    @SYNC_CALL
    def init_instance(self, adaptor_state, url, flags, session):

        # FIXME: eval flags!

        self._logger.info("init_instance %s" % url)

        if 'from_open' in adaptor_state and adaptor_state['from_open']:
            self.url = rsurl.Url(url)  # deep copy
            self.flags = flags
            self.session = session
            self.valid = False  # will be set by initialize
            self.cwdurl = rsurl.Url(adaptor_state["cwd"])
            self.cwd = self.cwdurl.path

            if rsumisc.url_is_relative(self.url):
                self.url = rsumisc.url_make_absolute(self.cwd, self.url)

        else:
            if rsumisc.url_is_relative(url):
                raise BadParameter("cannot interprete relative URL in this context ('%s')" % url)

            self.url = url
            self.flags = flags
            self.session = session
            self.valid = False  # will be set by initialize
            self.cwd = rsumisc.url_get_dirname(url)

            self.cwdurl = rsurl.Url(url)  # deep copy
            self.cwdurl.path = self.cwd

        self.initialize()
        return self.get_api()

    # ----------------------------------------------------------------
    #
    def initialize(self):

        if self.flags & CREATE_PARENTS:
            raise BadParameter("File creation not supported via HTTP(S)")

        elif self.flags & CREATE:
            raise BadParameter("File creation not supported via HTTP(S)")

        elif self.flags & WRITE:
            raise BadParameter("File write not supported via HTTP(S)")

        elif self.flags & READ:
            pass

        self.valid = True

    # ----------------------------------------------------------------
    #
    def finalize(self, kill=False):
        # nothing to do here
        pass

    # ----------------------------------------------------------------
    #
    @SYNC_CALL
    def get_url(self):
        return rsurl.Url(self.url)  # deep copy

    # ----------------------------------------------------------------
    #
    @SYNC_CALL
    def copy_self(self, tgt_in, flags):

        src = rsurl.Url(self.url)  # deep copy
        tgt = rsurl.Url(tgt_in)  # deep copy

        if tgt.scheme != 'file':
            raise BadParameter("Only file://localhost URLs are supported as copy targets.")

        if tgt.host != 'localhost':
            raise BadParameter("Only file://localhost URLs are supported as copy targets.")

        #if rsumisc.url_is_relative (src) : src = rsumisc.url_make_absolute (cwdurl, src)
        #if rsumisc.url_is_relative (tgt) : tgt = rsumisc.url_make_absolute (cwdurl, tgt)

        target = ""

        src_filename = os.path.basename(src.path)
        local_path = tgt.path
        if os.path.exists(tgt.path):
            if os.path.isfile(tgt.path):
                # fail if overwtrite flag is not set, otherwise copy
                if flags & OVERWRITE:
                    target = local_path
                else:
                    raise BadParameter("Local file '%s' exists." % local_path)

            elif os.path.isdir(tgt.path):
                # add source filename to target path
                target = os.path.join(local_path, src_filename)

                if os.path.exists(target):
                    if not flags & OVERWRITE:
                        raise BadParameter("Local file '%s' exists." % target)

        try:
            urllib.request.urlretrieve(str(src), target)
        except Exception as e:
            raise BadParameter("Couldn't copy %s to %s: %s" %
                                    (str(src), target, str(e))) from e

    # ----------------------------------------------------------------
    #
    @SYNC_CALL
    def is_dir_self(self):
        return False

    # ----------------------------------------------------------------
    #
    @SYNC_CALL
    def is_entry_self(self):
        return True

    # ----------------------------------------------------------------
    #
    @SYNC_CALL
    def is_link_self(self):
        return False

    # ----------------------------------------------------------------
    #
    @SYNC_CALL
    def is_file_self(self):
        return True




__author__    = "Andre Merzky"
__copyright__ = "Copyright 2012-2013, The SAGA Project"
__license__   = "MIT"


# namespace flags enum:
OVERWRITE      =    1
RECURSIVE      =    2
DEREFERENCE    =    4
CREATE         =    8
EXCLUSIVE      =   16
LOCK           =   32
CREATE_PARENTS =   64
# TRUNCATE     =  128 # reserved
# APPEND       =  256 # reserved
# READ         =  512 # reserved
# WRITE        = 1024 # reserved
# READ_WRITE   = 1536 # reserved
# BINARY       = 2048 # reserved








__author__    = "Andre Merzky"
__copyright__ = "Copyright 2012-2013, The SAGA Project"
__license__   = "MIT"


from .job         import Job
from .job         import Self
from .service     import Service
# from .description import Description







#!/usr/bin/env python
# -*- coding: utf-8 -*-
# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4

'''This examples shows how to copy a file to a remote
   host via SRM using a custom security context.

   If something doesn't work as expected, try to set
   SAGA_VERBOSE=3 in your environment before you run the
   script in order to get some debug output.

   If you think you have encountered a defect, please
   report it at: https://github.com/radical-cybertools/radical.saga/issues
'''

__author__    = "Mark Santcroos"
__copyright__ = "Copyright 2013, Mark Santcroos"
__license__   = "MIT"


import sys

import radical.saga as rs


def main():

    try:
        myfile = rs.filesystem.File('srm://tbn18.nikhef.nl/dpm/nikhef.nl/home/vlemed/mark/radical.saga/input.txt')
      # nofile = rs.filesystem.File('srm://tbn18.nikhef.nl/dpm/nikhef.nl/home/vlemed/mark/radical.saga/input.txt-non-exist')

      # mydir  = rs.filesystem.Directory('srm://tbn18.nikhef.nl/dpm/nikhef.nl/home/vlemed/mark/radical.saga')
      # nodir  = rs.filesystem.File('srm://tbn18.nikhef.nl/dpm/nikhef.nl/home/vlemed/mark/radical.saga-non-exists/input.txt-non-exist')

        print(myfile.get_size_self())
      # print myfile_non_exist.get_size_self()
      # print mydir.get_size()
      # print mydir_non_exist.get_size()

    except rs.SagaException as ex:
        print("An error occured during file operation: %s" % (str(ex)))
        sys.exit(-1)


if __name__ == "__main__":
    main()





__author__    = "Andre Merzky"
__copyright__ = "Copyright 2012-2013, The SAGA Project"
__license__   = "MIT"


# import logger
# 
# from   pty_process import *
# from   exception   import *
# from   misc        import *

from . import pty_shell





__author__    = "Andre Merzky, Ole Weidner"
__copyright__ = "Copyright 2012-2013, The SAGA Project"
__license__   = "MIT"


''' mock based job adaptor implementation '''

import re
import time
import threading as mt

import radical.utils as ru

from ...               import exceptions as rse
from ..                import base
from ..cpi             import SYNC_CALL
from ..cpi             import job as cpi
from ...               import job as api
from ...utils.job      import TransferDirectives
from ...utils          import pty_mock

from . import mock_wrapper


# ------------------------------------------------------------------------------
#
_PING_DELAY  = 60.0


# ------------------------------------------------------------------------------
#
class _job_state_monitor(mt.Thread):
    ''' thread that periodically monitors job states '''

    # --------------------------------------------------------------------------
    #
    def __init__(self, js, channel, rm, logger):

        self.js      = js
        self.channel = channel
        self.rm      = rm
        self.logger  = logger
        self.stop    = False
        self.events  = dict()

        super(_job_state_monitor, self).__init__()

        self.setDaemon(True)


    # --------------------------------------------------------------------------
    #
    def finalize(self):

        self.stop = True


    # --------------------------------------------------------------------------
    #
    def run(self):

        MONITOR_READ_TIMEOUT = 1.0   # check for stop signal now and then

        try:

            self.channel.run_async("MONITOR")

            while self.channel.alive():

                idx,out = self.channel.find(['\n'],timeout=MONITOR_READ_TIMEOUT)
                line    = out.strip()

                if not line:

                    # just a read timeout, i.e. an opportiunity to check for
                    if self.stop:
                        self.logger.debug("stop monitoring")
                        return
                    pass


                elif line == 'EXIT' or line == "Killed":
                    self.logger.error("monitor failed -- disable notifications")
                    return


                elif ':' not in line:
                    self.logger.warn("monitoring channel noise: %s" % line)


                else:
                    job_pid, state, data = line.split(':', 2)
                    job_id = "[%s]-[%s]" % (self.rm, job_pid)

                    state = self.js._adaptor.string_to_state(state)

                    try:
                        job = self.js.get_job(job_id, no_reconnect=True)

                        if not job:
                            # job not yet known -- keep event for later
                            if job_id not in self.events:
                                self.events[job_id] = list()
                            self.events[job_id].append(state)

                        else:

                            # check for previous events:
                            if job_id in self.events:
                                for event in self.events[job_id]:
                                    job._adaptor._set_state(event)
                                del(self.events[job_id])
                            job._adaptor._set_state(state)


                    except rse.DoesNotExist as e:
                        self.logger.error("event for unknown job '%s'" % job_id)

        except Exception as e:

            self.logger.error("Exception in job monitoring thread: %s" % e)
            self.logger.error("Cancel job monitoring for %s" % self.rm)


# ------------------------------------------------------------------------------
#
# strip white space from a string, and hex-decode the remaining characters.
# This must be applied to stdout/stderr data returned from the mock wrapper.
#
def _decode(data):

    code = ""

    for c in data:

        if c in ['0', '1', '2', '3', '4', '5', '6', '7', '8', '9',
                  'a', 'b', 'c', 'd', 'e', 'f']:
            code += c
        elif c not in [' ', '\n']:
            raise rse.BadParameter("Cannot decode '%s' in '%s'" % (c, data))

    return bytes.fromhex(code).decode('utf-8')


# ------------------------------------------------------------------------------
# the adaptor name
#
_ADAPTOR_NAME          = "radical.saga.adaptors.mock_job"
_ADAPTOR_SCHEMAS       = ["fork", "local", "ssh", "gsissh"]



# ------------------------------------------------------------------------------
# the adaptor capabilities & supported attributes
#
_ADAPTOR_CAPABILITIES  = {
    "jdes_attributes" : [api.NAME,
                          api.EXECUTABLE,
                          api.PRE_EXEC,
                          api.POST_EXEC,
                          api.ARGUMENTS,
                          api.ENVIRONMENT,
                          api.WORKING_DIRECTORY,
                          api.FILE_TRANSFER,
                          api.INPUT,
                          api.OUTPUT,
                          api.ERROR,
                          api.NAME,
                          api.WALL_TIME_LIMIT,
                          api.TOTAL_CPU_COUNT,
                          api.TOTAL_GPU_COUNT,
                          api.PROCESSES_PER_HOST,
                          api.SPMD_VARIATION,
                         ],
    "job_attributes"  : [api.EXIT_CODE,
                          api.EXECUTION_HOSTS,
                          api.CREATED,
                          api.STARTED,
                          api.FINISHED],
    "metrics"         : [api.STATE,
                          api.STATE_DETAIL],
    "contexts"        : {"ssh"     : "public/private keypair",
                          "x509"    : "X509 proxy for gsissh",
                          "userpass": "username/password pair for ssh"}
}

# ------------------------------------------------------------------------------
# the adaptor documentation
#
_ADAPTOR_DOC           = {
    "name"            : _ADAPTOR_NAME,
    "capabilities"    : _ADAPTOR_CAPABILITIES,
    "description"     : '''
        The Mock job adaptor. This adaptor uses the sh command line tools (sh,
        ssh, gsissh) to run local and remote jobs.  The adaptor expects the
        login mock on the target host to be POSIX compliant.  However, one can
        also specify a custom POSIX mock via the resource manager URL, like::

          js = rs.job.Service("ssh://remote.host.net/bin/sh")

        Note that custom mocks in many cases will find a different environment
        than the users default login mock!


        Known Limitations:
        ******************

          * number of system pty's are limited:  each job.service object bound
            to this adaptor will use 2 pairs of pty pipes.  Systems usually
            limit the number of available pty's to 1024 .. 4096.  Given that
            other process also use pty's , that gives a hard limit to the number
            of object instances which can be created concurrently.  Hitting the
            pty limit will cause the following error message (or similar)::

              NoSuccess: pty_allocation or process creation failed \
                         (ENOENT: no more ptys)

            This limitation comes from rs.utils.pty_process.  On Linux
            systems, the utilization of pty's can be monitored::

               echo "allocated pty's: `cat /proc/sys/kernel/pty/nr`"
               echo "available pty's: `cat /proc/sys/kernel/pty/max`"


          * number of ssh connections are limited: sshd's default configuration,
            which is in place on many systems, limits the number of concurrent
            ssh connections to 10 per user -- beyond that, connections are
            refused with the following error::

              NoSuccess: ssh_exchange_identification: \
                         Connection closed by remote host

            As the communication with the ssh channel is unbuffered, the
            dropping of the connection will likely cause this error message to
            be lost.  Instead, the adaptor will just see that the ssh connection
            disappeared, and will issue an error message similar to this one::

              NoSuccess: read from pty process failed \
                         (Could not read line - pty process died)


          * number of processes are limited: the creation of an job.service
            object will create one additional process on the local system, and
            two processes on the remote system (ssh daemon clone and a mock
            instance).  Each remote job will create three additional processes:
            two for the job instance itself (double fork), and an additional
            process which monitors the job for state changes etc.  Additional
            temporary processes may be needed as well.

            While marked as 'obsolete' by POSIX, the `ulimit` command is
            available on many systems, and reports the number of processes
            available per user (`ulimit -u`)

            On hitting process limits, the job creation will fail with an error
            similar to either of these::

              NoSuccess: failed to run job \
                         /bin/sh: fork: retry: Resource temporarily unavailable
              NoSuccess: failed to run job \
                         backend error

          * number of files are limited, as is disk space: the job.service will

            keep job state on the remote disk, in
            ``~/.radical/saga/adaptors/mock_job/``.
            Quota limitations may limit the number of files created there,
            and/or the total size of that directory.

            On quota or disk space limits, you may see error messages similar to
            the following ones::

              NoSuccess: read from pty process failed ([Errno 5] Quota exceeded)
              NoSuccess: read from pty process failed ([Errno 5] I/O error)
              NoSuccess: find from pty process [Thread-5] failed \
                         (Could not read - pty process died)



          * Other system limits (memory, CPU, selinux, accounting etc.) apply as
            usual.


          * thread safety: it is safe to create multiple:class:`job.Service`
            instances to the same target host at a time -- they should not
            interfere with each other, but ``list()`` will list jobs created by
            either instance (if those use the same target host user account).

            It is **not** safe to use the *same*:class:`job.Service` instance
            from multiple threads concurrently -- the communication on the I/O
            channel will likely get screwed up.  This limitation may be removed
            in future versions of the adaptor.  Non-concurrent (i.e. serialized)
            use should work as expected though.

        ''',
    "example"         : "examples/jobs/localjob.py",
    "schemas"         : {"fork"  :"use /bin/sh to run jobs",
                         "local" :"alias for fork://",
                         "ssh"   :"use ssh to run remote jobs",
                         "gsissh":"use gsissh to run remote jobs"}
}

# ------------------------------------------------------------------------------
# the adaptor info is used to register the adaptor with SAGA

_ADAPTOR_INFO = {
    "name"            : _ADAPTOR_NAME,
    "version"         : "v0.1",
    "schemas"         : _ADAPTOR_SCHEMAS,
    "capabilities"    : _ADAPTOR_CAPABILITIES,
    "cpis"            : [{
                                 "type" : "radical.saga.job.Service",
                                 "class": "MockJobService"
                         },{
                                 "type" : "radical.saga.job.Job",
                                 "class": "MockJob"
                         }]
}


# ------------------------------------------------------------------------------
# The adaptor class
class Adaptor(base.Base):
    '''
    This is the actual adaptor class, which gets loaded by SAGA (i.e. by the
    SAGA engine), and which registers the CPI implementation classes which
    provide the adaptor's functionality.
    '''


    # --------------------------------------------------------------------------
    #
    def __init__(self):

        base.Base.__init__(self, _ADAPTOR_INFO, expand_env=False)

        self.id_re = re.compile('^\[(.*)\]-\[(.*?)\]$')


    # --------------------------------------------------------------------------
    #
    def sanity_check(self):

        # FIXME: also check for gsissh

        pass


    # --------------------------------------------------------------------------
    #
    def parse_id(self, id):
        '''
        Split the id '[rm]-[pid]' in its parts, and return them.

        The callee makes sure that the ID is set and valid.
        '''

        match = self.id_re.match(id)

        if not match or len(match.groups()) != 2:
            raise rse.BadParameter("Cannot parse job id '%s'" % id)

        return(match.group(1), match.group(2))


    # --------------------------------------------------------------------------
    #
    def string_to_state(self, state_str):

        state_str = state_str.strip().lower()

        return {'new'      : api.NEW,
                'running'  : api.RUNNING,
                'suspended': api.SUSPENDED,
                'done'     : api.DONE,
                'failed'   : api.FAILED,
                'canceled' : api.CANCELED
               }.get(state_str, api.UNKNOWN)


    # --------------------------------------------------------------------------
    #
    def stage_input(self, mock, lock, jd):

        if not jd:
            return

        if jd.file_transfer is not None:

            td = TransferDirectives(jd.file_transfer)

            if td.in_append or td.out_append:
                raise rse.BadParameter('FT append(<</>>) not supported')

            if td.in_overwrite:
                for(local, remote) in td.in_overwrite:
                    source = local
                    target = remote
                    self._logger.info("stage %s to %s" % (source, target))

                    with lock:
                        mock.stage_to_remote(source, target)


    # --------------------------------------------------------------------------
    #
    def stage_output(self, mock, lock, jd):

        if not jd:
            return

        if jd.file_transfer is not None:
            td = TransferDirectives(jd.file_transfer)

            if td.out_append:
                raise rse.BadParameter('FileTransfer append not supported')

            if td.out_overwrite:
                for(local, remote) in td.out_overwrite:
                    source = remote
                    target = local
                    self._logger.info("stage %s to %s" % (source, target))

                    with lock:
                        mock.stage_from_remote(source, target)


# ------------------------------------------------------------------------------
#
class MockJobService(cpi.Service):

    # --------------------------------------------------------------------------
    #
    def __init__(self, api, adaptor):

        _cpi_base = super (MockJobService, self)
        _cpi_base.__init__(api, adaptor)

        self.opts = {}
        self.opts['mock'] = None  # default to login mock


    # --------------------------------------------------------------------------
    #
    def __del__(self):

        try   : self.close()
        except: pass


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def init_instance(self, adaptor_state, rm_url, session):
        '''
        Service instance constructor
        '''

        self.rm      = rm_url
        self.session = session
        self.jobs    = dict()
        self.njobs   = 0

        # Use `_set_session` method of the base class to set the session object.
        # `_set_session` and `get_session` methods are provided by `CPIBase`.
        self._set_session(session)

        # if the rm URL specifies a path, we interprete that as mock to run.
        # Otherwise, we default to running /bin/sh(for fork) or the user's
        # login mock(for ssh etc).
        if self.rm.path and self.rm.path != '/' and self.rm.path != '.':
            self.opts['mock'] = self.rm.path

        # create and initialize connection for starting jobs
        self.mock   = pty_mock.PTYMock(self.rm, self.get_session(),
                                           self._logger, cfg=self.opts)
        self.channel = pty_mock.PTYMock(self.rm, self.get_session(),
                                           self._logger, cfg=self.opts)
        self.initialize()

        # the monitoring thread - one per service instance.  We wait for
        # initialize to finish to make sure that the mock_wrapper is set
        # up...
        self.monitor = _job_state_monitor(js=self, rm=self.rm,
                                          channel=self.channel,
                                          logger=self._logger)
        self.monitor.start()

        return self.get_api()


    # --------------------------------------------------------------------------
    #
    def close(self):

        if self.mock:

            # FIXME: not sure if we should PURGE here -- that removes states
            # which might not be evaluated, yet.  Should we mark state
            # evaluation separately?
            #   cmd_state() { touch $DIR/purgeable; ... }
            # When should that be done?

            with self._mock_lock:
                # cancel scheduled `PING` request
                self._ping.cancel()

             #  self.mock.run_sync("PURGE", iomode=None)
                self.mock.run_async("QUIT")
                self.mock.finalize(kill_pty=True)
                self.mock = None

        if self.monitor:
            self.monitor.finalize()
            # we don't care about join, really


    # --------------------------------------------------------------------------
    #
    def initialize(self):

        # the underlying mock is not always ready to receive new commands,
        # e.g.,  when running an async command and still collecting output.
        # The driving code is usually aware of the mock's state - but we want
        # to send `PING` requests now and then to ensure the mock does not
        # time out, and the thread triggering that ping is *not* aware of the
        # mock state.  To avoid interfering with async calls, we lock any
        # interaction with the mock
        self._mock_lock = mt.RLock()

        # at regular intervals, run a ping toward the mock wrapper to avoid
        # timeouts kicking in
        # FIXME: configurable frequency
        self._ping = mt.Timer(_PING_DELAY, self._ping_cb)
        self._ping.start()


        # very first step: get the remote environment, and expand the config
        # settings
        with self._mock_lock:
            ret, out, err = self.mock.run_sync(' env')

        if ret != 0:
            raise rse.NoSuccess("env query failed(%s):(%s)(%s)"
                    % (ret, out, err))

        env = dict()
        for line in out.split('\n'):
            line = line.strip()
            if '=' in line:
                key, val = line.split('=', 1)
                env[key] = val

        cfg = self._adaptor._cfg

        # expand those config entries we want to use(where needed)
        self.notifications  = cfg.enable_notifications
        self.purge_on_start = cfg.purge_on_star
        self.base_workdir   = ru.expand_env(cfg.base_workdir, env)


        # start the mock, find its prompt.  If that is up and running, we can
        # bootstrap our wrapper script, and then run jobs etc.

        # -- now stage the mock wrapper script, and run it.  Once that is up
        # and running, we can requests job start / management operations via its
        # stdio.

        base = self.base_workdir

        with self._mock_lock:
            ret, out, _ = self.mock.run_sync(" mkdir -p %s" % base)

        if ret != 0:
            raise rse.NoSuccess("host setup failed(%s):(%s)" % (ret, out))

        # TODO: replace some constants in the script with values from config
        # files, such as 'timeout' or 'purge_on_quit' ...
        tgt = "%s/wrapper.sh" % base

        # lets check if we actually need to stage the wrapper script.  We need
        # an adaptor lock on this one.
        with self._adaptor._lock:

            with self._mock_lock:
                ret, out, _ = self.mock.run_sync(" test -f %s" % tgt)

            if ret != 0:
                # yep, need to stage...
                src = mock_wrapper._WRAPPER_SCRIPT
                src = src.replace('% (PURGE_ON_START)s',
                                  str(self.purge_on_start))

                # If the target directory begins with $HOME or ${HOME} then we
                # need to remove this since scp won't expand the variable and
                # the copy will end up attempting to copy the file to
                # /<path_to_home_dir>/$HOME/.....
                if tgt.startswith("$HOME") or tgt.startswith("${HOME}"):
                    tgt = tgt[tgt.find('/') + 1:]

                with self._mock_lock:
                    self.mock.write_to_remote(src, tgt)


        # ----------------------------------------------------------------------
        # we run the script.  In principle, we should set a new / different
        # prompt -- but, due to some strange and very unlikely coincidence, the
        # script has the same prompt as the previous mock... - go figure ;-)
        #
        # Note that we use 'exec' - so the script replaces the mock process.
        # Thus, when the script times out, the mock dies and the connection
        # drops -- that will free all associated resources, and allows for
        # a clean reconnect.
        # with self._mock_lock:
        #     ret, out, _ = self.mock.run_sync(" exec sh %s/wrapper.sh" % base)

        # Well, actually, we do not use exec, as that does not give us good
        # feedback on failures(the mock just quits) -- so we replace it with
        # this poor-man's version...
        with self._mock_lock:
            ret, out, _ = self.mock.run_sync(" /bin/sh %s/wrapper.sh %s"
                                             % (base, base))

        # mock_wrapper.sh will report its own PID -- we use that to sync prompt
        # detection, too.
        if ret != 0:
            raise rse.NoSuccess("failed to run bootstrap:(%s)(%s)" % (ret, out))

        id_pattern = re.compile("\s*PID:\s+(\d+)\s*$")
        id_match   = id_pattern.search(out)

        if not id_match:
            self.mock.run_async(" exit")
            self._logger.error("host bootstrap failed - no pid(%s)" % out)
            raise rse.NoSuccess  ("host bootstrap failed - no pid(%s)" % out)

        # we actually don't care much about the PID:-P

        self._logger.debug("got cmd prompt(%s)(%s)" % (ret, out.strip()))


        # ----------------------------------------------------------------------
        # now do the same for the monitoring mock
        ret, out, _ = self.channel.run_sync(" /bin/sh %s/wrapper.sh %s"
                                           % (base, base))

        # mock_wrapper.sh will report its own PID -- we use that to sync prompt
        # detection, too.
        if ret != 0:
            raise rse.NoSuccess("failed to run bootstrap:(%s)(%s)"
                               % (ret, out))

        id_pattern = re.compile("\s*PID:\s+(\d+)\s*$")
        id_match   = id_pattern.search(out)

        if not id_match:
            self.channel.run_async(" exit")
            self._logger.error("host bootstrap failed - no pid(%s)" % out)
            raise rse.NoSuccess  ("host bootstrap failed - no pid(%s)" % out)

        # we actually don't care much about the PID:-P

        self._logger.debug("got mon prompt(%s)(%s)" % (ret, out.strip()))


    # ----------------------------------------------------------------
    #
    def _ping_cb (self) :

        with self._mock_lock:

            _, out, _ = self.mock.run_sync('PING')
            assert('PONG' in out), out

            self._ping = mt.Timer(_PING_DELAY, self._ping_cb)
            self._ping.start()


    # --------------------------------------------------------------------------
    #
    #
    def _jd2cmd(self, jd):

        cmd = "true"

        if jd.attribute_exists(api.ENVIRONMENT):
            for e in jd.environment:
                cmd += " && export %s=%s"  % (e, jd.environment[e])

        if jd.attribute_exists(api.WORKING_DIRECTORY):
            cmd += " && mkdir -p %s && cd %s" \
                 % (jd.working_directory, jd.working_directory)

        if jd.attribute_exists(api.PRE_EXEC):
            for p in jd.pre_exec:
                cmd += " && %s 2>&1 >> $SAGA_PWD/log"  %  p

        cmd += " &&("
        cmd += " %s" % jd.executable

        if jd.attribute_exists(api.ARGUMENTS):
            for a in jd.arguments:
                cmd += " %s" % a

        cmd += " )"

        if jd.attribute_exists(api.INPUT):
            cmd += " <%s" % jd.input

        if jd.attribute_exists(api.OUTPUT):
            cmd += " 1>%s" % jd.output

        if jd.attribute_exists(api.ERROR):
            cmd += " 2>%s" % jd.error

        if jd.attribute_exists(api.POST_EXEC):
            for p in jd.post_exec:
                cmd += " && %s 2>&1 >> $SAGA_PWD/log"  %  p

        return cmd

    # --------------------------------------------------------------------------
    #
    #
    def _job_run(self, jd):
        '''
        runs a job on the wrapper via pty, and returns the job id
        '''

        # stage data, then run job
        self._adaptor.stage_input(self.mock, self._mock_lock, jd)

        # create command to run
        cmd = self._jd2cmd(jd)
        ret = 1
        out = ""

        run_cmd  = ""
        use_lrun = False

        # simple one-liners use RUN, otherwise LRUN
        if "\n" not in cmd:
            run_cmd = "RUN %s\n" % cmd
        else:
            use_lrun = True
            run_cmd  = "BULK\nLRUN\n%s\nLRUN_EOT\nBULK_RUN\n" % cmd

        run_cmd = run_cmd.replace("\\", "\\\\\\\\")  # hello MacOS

        with self._mock_lock:
            ret, out, _ = self.mock.run_sync(run_cmd)

        if ret != 0:
            raise rse.NoSuccess("failed to run Job '%s':(%s)(%s)"
                               % (cmd, ret, out))

        lines = [_f for _f in out.split("\n") if _f]
        self._logger.debug(lines)

        if len(lines) < 2:
            raise rse.NoSuccess("Failed to run job(%s)" % lines)

      # for i in range(0, len(lines)):
      #     print("%d: %s" % (i, lines[i]))

        if lines[-2] != "OK":
            raise rse.NoSuccess("Failed to run Job(%s)" % lines)

        # FIXME: verify format of returned pid(\d+)!
        pid    = lines[-1].strip()
        job_id = "[%s]-[%s]" % (self.rm, pid)

        self._logger.debug("started job %s" % job_id)

        self.njobs += 1

        # clean 'BULK COMPLETED message from lrun
        if use_lrun:
            with self._mock_lock:
                ret, out = self.mock.find_prompt()

            if ret != 0:
                raise rse.NoSuccess("failed to run multiline job '%s':(%s)(%s)"
                                   % (run_cmd, ret, out))
        return job_id


    # --------------------------------------------------------------------------
    #
    #
    def _job_get_stats(self, id):
        '''
        get the job stats from the wrapper mock
        '''

        rm, pid = self._adaptor.parse_id(id)

        with self._mock_lock:
            ret, out, _ = self.mock.run_sync("STATS %s\n" % pid)

        if ret != 0:
            raise rse.NoSuccess("failed to get job stats for '%s':(%s)(%s)"
                          % (id, ret, out))

        # the filter removes also empty lines from stdout/stderr.  Oh well...
        lines = [_f for _f in out.split("\n") if _f]
        self._logger.debug(lines)

        if lines[0] != "OK":
            raise rse.NoSuccess("failed to get valid job state for '%s'(%s)"
                               % (id, lines))
        ret           = dict()
        ret['STDOUT'] = ""
        ret['STDERR'] = ""
        in_stdout     = False
        in_stderr     = False

        for line in lines:

            if in_stdout:
                if 'END_STDOUT' in line:
                    in_stdout = False
                else:
                    ret['STDOUT'] += line
                continue

            if in_stderr:
                if 'END_STDERR' in line:
                    in_stderr = False
                else:
                    ret['STDERR'] += line
                continue

            if ':' in line:
                key, val = line.split(":", 1)
                ret[key.strip().lower()] = val.strip()
                continue

            if 'START_STDOUT' in line:
                in_stdout = True
                in_stderr = False
                continue

            if 'START_STDERR' in line:
                in_stderr = True
                in_stdout = False
                continue

          # import pprint
          # pprint.pprint(ret)

        return ret



    # --------------------------------------------------------------------------
    #
    # TODO: this should also fetch the(final) state, to safe a hop
    #
    def _job_get_exit_code(self, id):
        '''
        get the job's exit code from the wrapper mock
        '''

        rm, pid = self._adaptor.parse_id(id)

        with self._mock_lock:
            ret, out, _ = self.mock.run_sync("RESULT %s\n" % pid)

        if ret != 0:
            raise rse.NoSuccess ("failed to get exit code for '%s':(%s)(%s)"
                               % (id, ret, out))
            self._logger.warning("failed to get exit code for '%s':(%s)(%s)"
                               % (id, ret, out))
            return None

        lines = [_f for _f in out.split("\n") if _f]
        self._logger.debug(lines)

        if len(lines) == 3:
            # mock did not manage to do 'stty -echo'?
            del(lines[0])

        if len(lines) != 2:
            raise rse.NoSuccess("failed to get exit code for '%s':(%s)"
                               % (id, lines))

        if lines[0] != "OK":
            raise rse.NoSuccess("failed to get exit code for '%s'(%s)"
                               % (id, lines))

        exit_code = lines[1].strip()

        if not exit_code.isdigit():
            return None

        return int(exit_code)


    # --------------------------------------------------------------------------
    #
    # TODO: this should also cache state
    #
    def _job_suspend(self, id):

        rm, pid = self._adaptor.parse_id(id)

        with self._mock_lock:
            ret, out, _ = self.mock.run_sync("SUSPEND %s\n" % pid)

        if ret != 0:
            raise rse.NoSuccess("failed to suspend job '%s':(%s)(%s)"
                               % (id, ret, out))


    # --------------------------------------------------------------------------
    #
    # TODO: this should also cache state
    #
    def _job_resume(self, id):

        rm, pid = self._adaptor.parse_id(id)

        with self._mock_lock:
            ret, out, _ = self.mock.run_sync("RESUME %s\n" % pid)

        if ret != 0:
            raise rse.NoSuccess("failed to resume job '%s':(%s)(%s)"
                               % (id, ret, out))


    # --------------------------------------------------------------------------
    #
    # TODO: this should also fetch the(final) state, to safe a hop
    #
    def _job_cancel(self, id):

        rm, pid = self._adaptor.parse_id(id)

        with self._mock_lock:
            ret, out, err = self.mock.run_sync("CANCEL %s\n" % pid)

        if ret != 0:
            raise rse.NoSuccess("failed to cancel job '%s':(%s)(%s)(%s)"
                               % (id, ret, out, err))

        lines = [_f for _f in out.split("\n") if _f]
        self._logger.debug(lines)

        if len(lines) == 3:
            # mock did not manage to do 'stty -echo'?
            del(lines[0])

        if lines[0] != "OK":
            # this can happen if the cancel raced with some other final state
            # transition.  Lets check if the job is final by now - if it is, we
            # are good.
            if len(lines) > 1:
                state = lines[1].split(':')[1].strip()
                state = self._adaptor.string_to_state(state)
                if state in [api.DONE, api.FAILED, api.CANCELED]:
                    # nothing to do
                    return

            raise rse.NoSuccess("failed to cancel job '%s'(%s)" % (id, lines))


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def run_job(self, cmd, host):
        '''
        Implements adaptors.cpi.job.Service.run_job()
        '''

        if not cmd:
            raise rse.BadParameter._log(self._logger,
                    "run_job needs a command to run")

        if host and host != self.rm.host:
            raise rse.BadParameter._log(self._logger,
                    "Can only run jobs on %s, not on %s" % (self.rm.host, host))

        cmd_quoted = cmd.replace("'", "\\\\'")

        jd = api.Description()

        jd.executable = "/bin/sh"
        jd.arguments  = ["-c", "'%s'" % cmd_quoted]

        job = self.create_job(jd)
        job.run()

        return job


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def create_job(self, jd):

        # this dict is passed on to the job adaptor class -- use it to pass any
        # state information you need there.
        adaptor_state = {"job_service"    : self,
                         "job_description": jd,
                         "job_schema"     : self.rm.schema }

        return api.Job(_adaptor=self._adaptor, _adaptor_state=adaptor_state)


    # --------------------------------------------------------------------------
    @SYNC_CALL
    def get_url(self):

        return self.rm


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def list(self):

        # FIXME: this should also fetch job state and metadata, and cache those

        with self._mock_lock:
            ret, out, _ = self.mock.run_sync("LIST\n")

        if ret != 0:
            raise rse.NoSuccess("failed to list jobs:(%s)(%s)" % (ret, out))

        lines = [_f for _f in out.split("\n") if _f]
        self._logger.debug(lines)

        if lines[0] != "OK":
            raise rse.NoSuccess("failed to list jobs(%s)" % (lines))

        del lines[0]
        job_ids = list()

        for line in lines:

            try:
                pid    = line.strip()
                job_id = "[%s]-[%s]" % (self.rm, pid)
                job_ids.append(job_id)

            except Exception as e:
                self._logger.debug("Ignore ill-formatted job id(%s)(%s)"
                                  % (line, e))
                continue

        return job_ids


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def get_job(self, job_id, no_reconnect=False):

        if job_id in self.jobs:
            # no need to reconnect
            return self.jobs[job_id]

        if no_reconnect:
            return None

        known_jobs = self.list()

        if job_id not in known_jobs:
            # can't reconnect
            raise rse.BadParameter._log(self._logger, "job id '%s' unknown"
                                       % job_id)

        # this dict is passed on to the job adaptor class -- use it to pass any
        # state information you need there.
        adaptor_state = {"job_service"    : self,
                         "job_id"         : job_id,
                         "job_schema"     : self.rm.schema }

        return api.Job(_adaptor=self._adaptor, _adaptor_state=adaptor_state)


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def container_run(self, jobs):
        '''
        From all the job descriptions in the container, build a bulk, and submit
        as async.  The read whaterver the wrapper returns, and sort through the
        messages, assigning job IDs etc.
        '''

        # FIXME: this just assumes that all tasks are job creation tasks --
        #        which is not necessarily true...

        self._logger.debug("container run: %s"  %  str(jobs))

        bulk = "BULK\n"

        for job in jobs:
            cmd   = self._jd2cmd(job.description)
            bulk += "RUN %s\n" % cmd

        # ------------------------------------------------------------
        # stage input data
        # FIXME: this is now blocking the run() method.  Ideally, this activity
        # should be passed to a data manager thread/process/service.
        for job in jobs:
            self._adaptor.stage_input(self.mock, self._mock_lock,
                                      job.description)
        # ------------------------------------------------------------

        bulk += "BULK_RUN\n"
        with self._mock_lock:

            self.mock.run_async(bulk)

            for job in jobs:

                ret, out = self.mock.find_prompt()

                if ret != 0:
                    job._adaptor._set_state(api.FAILED)
                    job._adaptor._exception = rse.NoSuccess \
                           ("failed to run job:(%s)(%s)" % (ret, out))
                    continue

                lines = [_f for _f in out.split("\n") if _f]

                if len(lines) < 2:
                    job._adaptor._set_state(api.FAILED)
                    job._adaptor._exception = rse.NoSuccess \
                           ("failed to run job:(%s)(%s)" % (ret, out))
                    continue

                if lines[-2] != "OK":
                    job._adaptor._set_state(api.FAILED)
                    job._adaptor._exception = rse.NoSuccess \
                           ("failed to run job:(%s)(%s)" % (ret, out))
                    continue

                # FIXME: verify format of returned pid(\d+)!
                pid    = lines[-1].strip()
                job_id = "[%s]-[%s]" % (self.rm, pid)

                self._logger.debug("started job %s" % job_id)

                self.njobs += 1

                # FIXME: at this point we need to make sure that we actually created
                # the job.  Well, we should make sure of this *before* we run it.
                # But, actually, the container sorter should have done that already?
                # Check!
                job._adaptor._id = job_id

            # we also need to find the output of the bulk op itself
            ret, out = self.mock.find_prompt()

            if ret != 0:
                self._logger.error("failed to run(parts of ) bulk jobs:(%s)(%s)"
                                  % (ret, out))
                return

            lines = [_f for _f in out.split("\n") if _f]

            if len(lines) < 2:
                self._logger.error("no status of bulk job submission:(%s)(%s)"
                                  % (ret, out))
                return

            if lines[-2] != "OK":
                self._logger.error("failed to run(parts of ) bulk jobs:(%s)(%s)"
                                  % (ret, out))
                return


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def container_wait(self, jobs, mode, timeout):

        # FIXME: this just assumes that all tasks are job wait tasks --
        #        which is not necessarily true...
        # FIXME: we ignore the job wait mode(ALL/ANY), and always wait for all
        #        jobs...

        self._logger.debug("container wait: %s"  %  str(jobs))

        bulk = "BULK\n"

        for job in jobs:
          # print(type(job))
          # print(type(job._adaptor))

            if not isinstance(job._adaptor, MockJob):
                # this is not a job created by this adaptor.  Its probably
                # a task for a job operation where the job is owned by this
                # adaptor(FIXME: check).  Fall back to non-container wait.
                # FIXME: timeout handling is wrong
                job.wait(timeout)
            else:
                rm, pid = self._adaptor.parse_id(job.id)
                bulk   += "WAIT %s\n" % pid

        bulk += "BULK_RUN\n"

        with self._mock_lock:
            self.mock.run_async(bulk)

            for job in jobs:

                ret, out = self.mock.find_prompt()

                if ret != 0:
                    job._adaptor._set_state(api.FAILED)
                    job._adaptor._exception = rse.NoSuccess \
                           ("failed to wait for job:(%s)(%s)" % (ret, out))
                    continue

                lines = [_f for _f in out.split("\n") if _f]

                if len(lines) < 2:
                    job._adaptor._set_state(api.FAILED)
                    job._adaptor._exception = rse.NoSuccess \
                           ("failed to wait for job:(%s)(%s)" % (ret, out))
                    continue

                if lines[-2] != "OK":
                    job._adaptor._set_state(api.FAILED)
                    job._adaptor._exception = rse.NoSuccess \
                           ("failed to wait for job:(%s)(%s)" % (ret, out))
                    continue

            # we also need to find the output of the bulk op itself
            ret, out = self.mock.find_prompt()

            if ret != 0:
                self._logger.error("failed to wait for(part of) bulk job:(%s)(%s)"
                                  % (ret, out))
                return

            lines = [_f for _f in out.split("\n") if _f]

            if len(lines) < 2:
                self._logger.error("no status of bulk job wait:(%s)(%s)"
                                  % (ret, out))
                return

            if lines[-2] != "OK":
                self._logger.error("failed to wait for part of bulk job:(%s)(%s)"
                                  % (ret, out))
                return


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def container_cancel(self, jobs, timeout):

        self._logger.debug("container cancel: %s [%s]"  % (str(jobs), timeout))

        bulk = "BULK\n"

        for job in jobs:
            rm, pid = self._adaptor.parse_id(job.id)
            bulk   += "CANCEL %s\n" % pid

        bulk += "BULK_RUN\n"

        with self._mock_lock:
            self.mock.run_async(bulk)

            for job in jobs:

                ret, out = self.mock.find_prompt()

                if ret != 0:
                    job._adaptor._set_state(api.FAILED)
                    job._adaptor._exception = rse.NoSuccess \
                           ("failed to cancel job:(%s)(%s)" % (ret, out))
                    continue

                lines = [_f for _f in out.split("\n") if _f]

                if len(lines) < 2:
                    job._adaptor._set_state(api.FAILED)
                    job._adaptor._exception = rse.NoSuccess \
                           ("failed to cancel job:(%s)(%s)" % (ret, out))
                    continue

                if lines[-2] != "OK":
                    job._adaptor._set_state(api.FAILED)
                    job._adaptor._exception = rse.NoSuccess \
                           ("failed to cancel job:(%s)(%s)" % (ret, out))
                    continue

            # we also need to find the output of the bulk op itself
            ret, out = self.mock.find_prompt()

            if ret != 0:
                self._logger.error("failed to cancel part of bulk job:(%s)(%s)"
                                  % (ret, out))
                return

            lines = [_f for _f in out.split("\n") if _f]

            if len(lines) < 2:
                self._logger.error("no status of bulk job cancel:(%s)(%s)"
                                  % (ret, out))
                return

            if lines[-2] != "OK":
                self._logger.error("failed to cancel part of bulk job:(%s)(%s)"
                                  % (ret, out))
                return


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def container_get_states(self, jobs):

        self._logger.debug("container get_state: %s"  %  str(jobs))

        bulk   = "BULK\n"
        states = []

        for job in jobs:

          # print(job)
          # job._attributes_dump()

            rm, pid = self._adaptor.parse_id(job.id)
            bulk   += "STATE %s\n" % pid

        bulk += "BULK_RUN\n"

        with self._mock_lock:
            self.mock.run_async(bulk)

            for job in jobs:

                ret, out = self.mock.find_prompt()

                if ret != 0:
                    job._adaptor._set_state(api.FAILED)
                    job._adaptor._exception = rse.NoSuccess \
                           ("failed to get job state:(%s)(%s)" % (ret, out))
                    continue

                lines = [_f for _f in out.split("\n") if _f]

                if len(lines) < 2:
                    job._adaptor._set_state(api.FAILED)
                    job._adaptor._exception = rse.NoSuccess \
                           ("failed to get job state:(%s)(%s)" % (ret, out))
                    continue

                if lines[-2] != "OK":
                    job._adaptor._set_state(api.FAILED)
                    job._adaptor._exception = rse.NoSuccess \
                           ("failed to get job state:(%s)(%s)" % (ret, out))
                    continue

                state = self._adaptor.string_to_state(lines[-1])

                job._adaptor._update_state(state)
                states.append(state)


            # we also need to find the output of the bulk op itself
            ret, out = self.mock.find_prompt()

            if ret != 0:
                self._logger.error("no state for part of bulk job:(%s)(%s)"
                                  % (ret, out))
                return

            lines = [_f for _f in out.split("\n") if _f]

            if len(lines) < 2:
                self._logger.error("Cannot eval status of bulk job:(%s)(%s)"
                                  % (ret, out))
                return

            if lines[-2] != "OK":
                self._logger.error("no state for part of bulk job:(%s)(%s)"
                                  % (ret, out))
                return

        return states


# ------------------------------------------------------------------------------
#
class MockJob(cpi.Job):

    # --------------------------------------------------------------------------
    #
    def __init__(self, api, adaptor):

        _cpi_base = super (MockJob, self)
        _cpi_base.__init__(api, adaptor)


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def init_instance(self, job_info):

        if 'job_description' in job_info:
            # comes from job.service.create_job()
            self.js = job_info["job_service"]
            self.jd = job_info["job_description"]

            # the js is responsible for job bulk operations -- which
            # for jobs only work for run()
            self._container       = self.js

            # initialize job attribute values
            self._id              = None
            self._name            = self.jd.get(api.NAME)
            self._log             = list()
            self._state           = None
            self._exit_code       = None
            self._exception       = None
            self._created         = time.time()
            self._name            = self.jd.name
            self._started         = None
            self._finished        = None

            self._set_state(api.NEW)

        elif 'job_id' in job_info:
            # initialize job attribute values
            self.js               = job_info["job_service"]
            self.jd               = None
            self._id              = job_info['job_id']
            self._name            = job_info.get('job_name')
            self._log             = list()
            self._state           = None
            self._exit_code       = None
            self._exception       = None
            self._created         = None
            self._name            = None
            self._started         = None
            self._finished        = None

        else:
            # don't know what to do...
            raise rse.BadParameter("insufficient info for job creation")

        if self._created: self._created = float(self._created)

        return self.get_api()


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def get_description(self):

        return self.jd


    # --------------------------------------------------------------------------
    #
    def _update_state(self, state):

        old_state = self._state

        if state == api.DONE and \
            old_state not in [api.DONE, api.FAILED, api.CANCELED]:

            # stage output data
            # FIXME: _update_state blocks until data are staged.
            #        That should not happen.
            self._adaptor.stage_output(self.js.mock, self.js._mock_lock,
                                       self.jd)

        # files are staged -- update state, and report to application
        self._state = state
        if self._state != old_state:
            self._api()._attributes_i_set('state', self._state, self._api()._UP)


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def get_state(self):
        '''
        Implements adaptors.cpi.job.Job.get_state()
        '''

        # may not yet have backend representation, state is 'NEW'
        if self._id is None:
            return self._state

        stats = self.js._job_get_stats(self._id)

        if 'start' in stats: self._started  = stats['start']
        if 'stop'  in stats: self._finished = stats['stop']

        if self._started : self._started  = float(self._started)
        if self._finished: self._finished = float(self._finished)

        # no need to re-check final states
        if self._state == api.DONE      or \
           self._state == api.FAILED    or \
           self._state == api.CANCELED    :
                return self._state

        if 'state' not in stats:
            raise rse.NoSuccess("failed to get job state for '%s':(%s)"
                               % (self._id, stats))

        if self._exit_code is None and stats.get('ecode'):
            self._exit_code = int(stats['ecode'])

        self._update_state(self._adaptor.string_to_state(stats['state']))

        return self._state


    # --------------------------------------------------------------------------
    #
    def _set_state(self, state):

        old_state = self._state

        # on state changes, trigger notifications
        if old_state != state:
            self._state  = state
            self._api()._attributes_i_set('state', self._state, self._api()._UP)

        return self._state


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def get_created(self):

        # no need to refresh stats -- this is set locally
        return self._created


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def get_started(self):

        self.get_state()  # refresh stats
        return self._started


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def get_finished(self):

        self.get_state()  # refresh stats
        return self._finished


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def get_stdout(self):

        state = self.get_state() # refresh stats

        if state == api.NEW      or \
            state == api.PENDING :
            raise rse.IncorrectState \
                   ("Job output is only available after the job started")

        if not self._id:
            raise rse.IncorrectState \
                   ("Job output is only available after the job started")

        rm, pid     = self._adaptor.parse_id(self._id)
        ret, out, _ = self.js.mock.run_sync("STDOUT %s\n" % pid)

        if ret != 0:
            raise rse.NoSuccess \
                   ("failed to get job stdout for '%s':(%s)(%s)"
                     % (self._id, ret, out))

        lines = [_f for _f in out.split("\n") if _f]

        if lines[0] != "OK":
            raise rse.NoSuccess \
                   ("failed to get valid job stdout for '%s'(%s)"
                     % (self._id, lines))

        return _decode('\n'.join(lines[1:]))


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def get_stderr(self):

        state = self.get_state() # refresh stats

        if state == api.NEW      or \
            state == api.PENDING :
            raise rse.IncorrectState \
                   ("Job output is only available after the job started")

        if not self._id:
            raise rse.IncorrectState \
                   ("Job output is only available after the job started")

        rm, pid     = self._adaptor.parse_id(self._id)
        ret, out, _ = self.js.mock.run_sync("STDERR %s\n" % pid)

        if ret != 0:
            raise rse.NoSuccess \
                   ("failed to get job stderr for '%s':(%s)(%s)"
                     % (self._id, ret, out))

        lines = [_f for _f in out.split("\n") if _f]

        if lines[0] != "OK":
            raise rse.NoSuccess \
                   ("failed to get valid job stderr for '%s'(%s)"
                     % (self._id, lines))

        return _decode('\n'.join(lines[1:]))


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def get_log(self):

        state = self.get_state()  # refresh stats

        if state == api.NEW      or \
           state == api.PENDING :
            raise rse.IncorrectState("Job output is not yet available")

        if not self._id:
            raise rse.IncorrectState \
                   ("Job output is only available after the job started")

        rm, pid     = self._adaptor.parse_id(self._id)
        ret, out, _ = self.js.mock.run_sync("LOG %s\n" % pid)

        if ret != 0:
            raise rse.NoSuccess \
                   ("failed to get job log for '%s':(%s)(%s)"
                     % (self._id, ret, out))

        lines = [_f for _f in out.split("\n") if _f]

        if lines[0] != "OK":
            raise rse.NoSuccess \
                   ("failed to get valid job stderr for '%s'(%s)"
                     % (self._id, lines))

        ret  = '\n'.join(self._log)  # pre-pend all local log messages
        ret += _decode('\n'.join(lines[1:]))

        return ret


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def get_service_url(self):

        if not self.js:
            raise rse.IncorrectState("Job Service URL unknown")
        else:
            return self.js.get_url()


    # --------------------------------------------------------------------------
    #
    # TODO: this should also fetch the(final) state, to safe a hop
    # TODO: implement via notifications
    #
    @SYNC_CALL
    def wait(self, timeout):
        '''
        A call to the mock to do the WAIT would block the mock for any
        other interactions.  In particular, it would practically kill it if the
        Wait waits forever...

        So we implement the wait via a state pull.  The *real* solution is, of
        course, to implement state notifications, and wait for such
        a notification to arrive within timeout seconds...
        '''

        time_start = time.time()
        time_now   = time_start

        while True:

            state = self.get_state()

            if state in [api.DONE, api.FAILED, api.CANCELED]:
                return True

            # avoid busy poll
            # FIXME: self-tune by checking call latency
            time.sleep(0.1)

            # check if we hit timeout
            if timeout >= 0:
                time_now = time.time()
                if time_now - time_start > timeout:
                    return False

    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def get_id(self):

        return self._id


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def get_name(self):

        return self._name


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def get_exit_code(self):

        if self._exit_code is not None:
            return self._exit_code

        if self.get_state() not in [api.DONE, api.FAILED, api.CANCELED]:
            raise rse.IncorrectState \
                   ("Cannot get exit code, job is not in final state")

        self._exit_code = self.js._job_get_exit_code(self._id)

        return self._exit_code


    # --------------------------------------------------------------------------
    #
    # TODO: the values below should be fetched with every get_state...
    #
    @SYNC_CALL
    def get_execution_hosts(self):

        return [self.js.get_url().host]


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def run(self):

        self._id = self.js._job_run(self.jd)
        self.js.jobs[self._id] = self._api()

        self._set_state(api.RUNNING)


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def suspend(self):

        if self.get_state() != api.RUNNING:
            raise rse.IncorrectState("Cannot suspend, job is not RUNNING")

        self.js._job_suspend(self._id)


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def resume(self):

        if self.get_state() != api.SUSPENDED:
            raise rse.IncorrectState("Cannot resume, job is not SUSPENDED")

        self.js._job_resume(self._id)


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def cancel(self, timeout):

        if self.get_state() not in [api.RUNNING,  api.SUSPENDED,
                                    api.CANCELED, api.DONE,
                                    api.FAILED]:
            raise rse.IncorrectState("Cannot cancel, job is not running")

        if self._state in [api.CANCELED, api.DONE, api.FAILED]:
            self._set_state(api.CANCELED)
            return

        self.js._job_cancel(self._id)


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def re_raise(self):
        # nothing to do here actually, as run() is synchronous...
        return self._exception


# ------------------------------------------------------------------------------




#!/usr/bin/env python

__author__    = "Andre Merzky, Ole Weidner"
__copyright__ = "Copyright 2012-2013, The SAGA Project"
__license__   = "MIT"


import os
import sys
import getpass

import radical.saga as rs


REMOTE_HOST = "localhost"

def main():
    try:
        # Your ssh identity on the remote machine
        ctx = rs.Context("ssh")
        ctx.user_id = "oweidner"

        session = rs.Session()
      # session.add_context(ctx)

        # Create a job service object that represent the local machine.
        # The keyword 'fork://' in the url scheme triggers the 'shell' adaptor
        # which can execute jobs on the local machine as well as on a remote
        # machine via "ssh://hostname".
        js = rs.job.Service("ssh://%s" % REMOTE_HOST, session=session)

        # describe our job
        jd = rs.job.Description()

        # Next, we describe the job we want to run. A complete set of job
        # description attributes can be found in the API documentation.
        jd.environment     = {'MYOUTPUT':'"Hello from SAGA"'}
        jd.executable      = '/bin/echo'
        jd.arguments       = ['$MYOUTPUT']
        jd.output          = "/tmp/mysagajob-%s.stdout" % getpass.getuser()
        jd.error           = "/tmp/mysagajob-%s.stderr" % getpass.getuser()

        # Create a new job from the job description. The initial state of
        # the job is 'New'.
        myjob = js.create_job(jd)

        # Check our job's id and state
        print("Job ID    : %s" % (myjob.id))
        print("Job State : %s" % (myjob.state))

        print("\n...starting job...\n")

        # Now we can start our job.
        myjob.run()

        print("Job ID    : %s" % (myjob.id))
        print("Job State : %s" % (myjob.state))

        print("\n...waiting for job...\n")
        # wait for the job to either finish or fail
        myjob.wait()

        print("Job State : %s" % (myjob.state))
        print("Exitcode  : %s" % (myjob.exit_code))

        outfilesource = 'sftp://%s/tmp/mysagajob-%s.stdout' % (REMOTE_HOST, getpass.getuser())
        outfiletarget = "file://%s/" % os.getcwd()
        out = rs.filesystem.File(outfilesource, session=session)
        out.copy(outfiletarget)

        print("Staged out %s to %s (size: %s bytes)" % (outfilesource, outfiletarget, out.get_size()))


        return 0

    except rs.SagaException as ex:
        # Catch all saga exceptions
        print("An exception occured: (%s) %s " % (ex.type, (str(ex))))
        # Trace back the exception. That can be helpful for debugging.
        print(" \n*** Backtrace:\n %s" % ex.traceback)
        return -1


if __name__ == "__main__":
    sys.exit(main())




__author__    = "Andre Merzky, Ole Weidner, Alexander Grill"
__copyright__ = "Copyright 2012-2013, The SAGA Project"
__license__   = "MIT"


import radical.utils               as ru
import radical.utils.signatures    as rus

from .constants  import *
from ..constants import SYNC, ASYNC, TASK
from ..adaptors  import base       as sab
from ..namespace import entry      as nsentry

from .. import session             as ss
from .. import task                as st


# ------------------------------------------------------------------------------
#
class File (nsentry.Entry) :
    """
    Represents a local or remote file.

    The saga.filesystem.File class represents, as the name indicates,
    a file on some (local or remote) filesystem.  That class offers
    a number of operations on that file, such as copy, move and remove::

        # get a file handle
        file = saga.filesystem.File("sftp://localhost/tmp/data/data.bin")

        # copy the file
        file.copy ("sftp://localhost/tmp/data/data.bak")

        # move the file
        file.move ("sftp://localhost/tmp/data/data.new")
    """

    # --------------------------------------------------------------------------
    #
    @rus.takes   ('File',
                  rus.optional ((ru.Url, str)),
                  rus.optional (int, rus.nothing),
                  rus.optional (ss.Session),
                  rus.optional (sab.Base),
                  rus.optional (dict),
                  rus.optional (rus.one_of (SYNC, ASYNC, TASK)))
    @rus.returns (rus.nothing)
    def __init__ (self, url=None, flags=READ, session=None,
                  _adaptor=None, _adaptor_state={}, _ttype=None) :
        """
        __init__(url, flags=READ, session)

        Construct a new file object

        :param url:     Url of the (remote) file
        :type  url:     :class:`saga.Url`

        :fgs:   :ref:`filesystemflags`
        :param session: :class:`saga.Session`

        The specified file is expected to exist -- otherwise a DoesNotExist
        exception is raised.  Also, the URL must point to a file (not to
        a directory), otherwise a BadParameter exception is raised.

        Example::

            # get a file handle
            file = saga.filesystem.File("sftp://localhost/tmp/data/data.bin")

            # print the file's size
            print(file.get_size ())
        """

        # param checks
        if  not flags : flags = 0
        url = ru.Url (url)

        if  not url.schema :
            url.schema = 'file'

        if  not url.host :
            url.host = 'localhost'

        self._nsentry = super  (File, self)
        self._nsentry.__init__ (url, flags, session,
                                _adaptor, _adaptor_state, _ttype=_ttype)

    # --------------------------------------------------------------------------
    #
    @classmethod
    @rus.takes   ('File',
                  rus.optional ((ru.Url, str)),
                  rus.optional (int, rus.nothing),
                  rus.optional (ss.Session),
                  rus.optional (rus.one_of (SYNC, ASYNC, TASK)))
    @rus.returns (st.Task)
    def create (cls, url=None, flags=READ, session=None, ttype=None) :
        """
        create(url, flags, session)

        url:       saga.Url
        flags:     saga.replica.flags enum
        session:   saga.Session
        ttype:     saga.task.type enum
        ret:       saga.Task
        """
        if  not flags : flags = 0
        _nsentry = super (File, cls)
        return _nsentry.create (url, flags, session, ttype=ttype)


    # ----------------------------------------------------------------
    #
    # filesystem methods
    #
    # --------------------------------------------------------------------------
    #
    @rus.takes   ('File',
                  rus.optional (rus.one_of (SYNC, ASYNC, TASK)))
    @rus.returns ((bool, st.Task))
    def is_file (self, ttype=None) :
        """
        is_file()

        Returns `True` if instance points to a file, `False` otherwise.
        """
        return self._adaptor.is_file_self (ttype=ttype)


    # --------------------------------------------------------------------------
    #
    @rus.takes   ('File',
                  rus.optional (rus.one_of (SYNC, ASYNC, TASK)))
    @rus.returns ((int, st.Task))
    def get_size (self, ttype=None) :
        '''
        get_size()

        Returns the size (in bytes) of a file.

           Example::

               # get a file handle
               file = saga.filesystem.File("sftp://localhost/tmp/data/data.bin")

               # print the file's size
               print(file.get_size ())

        '''
        return self._adaptor.get_size_self (ttype=ttype)


    # --------------------------------------------------------------------------
    #
    @rus.takes   ('File',
                  rus.optional (int),
                  rus.optional (rus.one_of (SYNC, ASYNC, TASK)))
    @rus.returns ((str, st.Task))
    def read     (self, size=None, ttype=None) :
        '''
        size :    int
        ttype:    saga.task.type enum
        ret:      string / bytearray / saga.Task
        '''
        return self._adaptor.read (size, ttype=ttype)

    # --------------------------------------------------------------------------
    #
    @rus.takes   ('File',
                  rus.optional (bool))
    @rus.returns (st.Task)
    def close     (self, kill=True, ttype=None) :
        '''
        kill :    bool
        ttype:    saga.task.type enum
        ret:      string / bytearray / saga.Task
        '''
        return self._adaptor.close ()

    # --------------------------------------------------------------------------
    #
    @rus.takes   ('File',
                  str,
                  rus.optional (rus.one_of (SYNC, ASYNC, TASK)))
    @rus.returns ((int, st.Task))
    def write    (self, data, ttype=None) :
        '''
        data :    string / bytearray
        ttype:    saga.task.type enum
        ret:      int / saga.Task
        '''
        return self._adaptor.write (data, ttype=ttype)


    # --------------------------------------------------------------------------
    #
    @rus.takes   ('File',
                  int,
                  rus.optional (rus.one_of (START, CURRENT, END )),
                  rus.optional (rus.one_of (SYNC,  ASYNC,   TASK)))
    @rus.returns ((int, st.Task))
    def seek     (self, offset, whence=START, ttype=None) :
        '''
        offset:   int
        whence:   seek_mode enum
        ttype:    saga.task.type enum
        ret:      int / saga.Task
        '''
        return self._adaptor.seek (offset, whence, ttype=ttype)


    # --------------------------------------------------------------------------
    #
    @rus.takes   ('File',
                  rus.list_of  (rus.tuple_of (int)),
                  rus.optional (rus.one_of (SYNC, ASYNC, TASK)))
    @rus.returns ((str, st.Task))
    def read_v   (self, iovecs, ttype=None) :
        '''
        iovecs:   list [tuple (int, int)]
        ttype:    saga.task.type enum
        ret:      list [bytearray] / saga.Task
        '''
        return self._adaptor.read_v (iovecs, ttype=ttype)


    # --------------------------------------------------------------------------
    #
    @rus.takes   ('File',
                  rus.list_of  (rus.tuple_of ((int, str))),
                  rus.optional (rus.one_of (SYNC, ASYNC, TASK)))
    @rus.returns ((rus.list_of (int), st.Task))
    def write_v (self, data, ttype=None) :
        '''
        data:     list [tuple (int, string / bytearray)]
        ttype:    saga.task.type enum
        ret:      list [int] / saga.Task
        '''
        return self._adaptor.write_v (data, ttype=ttype)


    # --------------------------------------------------------------------------
    #
    @rus.takes   ('File',
                  str,
                  rus.optional (rus.one_of (SYNC, ASYNC, TASK)))
    @rus.returns ((int, st.Task))
    def size_p (self, pattern, ttype=None) :
        '''
        pattern:  string
        ttype:    saga.task.type enum
        ret:      int / saga.Task
        '''
        return self._adaptor.size_p (pattern, ttype=ttype)


    # --------------------------------------------------------------------------
    #
    @rus.takes   ('File',
                  str,
                  rus.optional (rus.one_of (SYNC, ASYNC, TASK)))
    @rus.returns ((str, st.Task))
    def read_p (self, pattern, ttype=None) :
        '''
        pattern:  string
        ttype:    saga.task.type enum
        ret:      string / bytearray / saga.Task
        '''
        return self._adaptor.read_p (pattern, ttype=ttype)


    # --------------------------------------------------------------------------
    #
    @rus.takes   ('File',
                  str,
                  str,
                  rus.optional (rus.one_of (SYNC, ASYNC, TASK)))
    @rus.returns ((int, st.Task))
    def write_p (self, pattern, data, ttype=None) :
        '''
        pattern:  string
        data:     string / bytearray
        ttype:    saga.task.type enum
        ret:      int / saga.Task
        '''
        return self._adaptor.write_p (pattern, data, ttype=ttype)


    # --------------------------------------------------------------------------
    #
    @rus.takes   ('File',
                  rus.optional (rus.one_of (SYNC, ASYNC, TASK)))
    @rus.returns ((rus.list_of (str), st.Task))
    def modes_e (self, ttype=None) :
        '''
        ttype:    saga.task.type enum
        ret:      list [string] / saga.Task
        '''
        return self._adaptor.modes_e (ttype=ttype)


    # --------------------------------------------------------------------------
    #
    @rus.takes   ('File',
                  str,
                  str,
                  rus.optional (rus.one_of (SYNC, ASYNC, TASK)))
    @rus.returns ((int, st.Task))
    def size_e (self, emode, spec, ttype=None) :
        '''
        emode:    string
        spec:     string
        ttype:    saga.task.type enum
        ret:      int / saga.Task
        '''
        return self._adaptor.size_e (emode, spec, ttype=ttype)


    # --------------------------------------------------------------------------
    #
    @rus.takes   ('File',
                  str,
                  str,
                  rus.optional (rus.one_of (SYNC, ASYNC, TASK)))
    @rus.returns ((str, st.Task))
    def read_e (self, emode, spec, ttype=None) :
        '''
        emode:    string
        spec:     string
        ttype:    saga.task.type enum
        ret:      bytearray / saga.Task
        '''
        return self._adaptor.read_e (emode, spec, ttype=ttype)


    # --------------------------------------------------------------------------
    #
    @rus.takes   ('File',
                  str,
                  str,
                  str,
                  rus.optional (rus.one_of (SYNC, ASYNC, TASK)))
    @rus.returns ((int, st.Task))
    def write_e (self, emode, spec, data, ttype=None) :
        '''
        emode:    string
        spec:     string
        data:     string / bytearray
        ttype:    saga.task.type enum
        ret:      int / saga.Task
        '''
        return self._adaptor.read_e (emode, spec, data, ttype=ttype)


    size    = property (get_size)  # int
    modes_e = property (modes_e)   # list [string]








__author__    = "Andre Merzky"
__copyright__ = "Copyright 2013, The SAGA Project"
__license__   = "MIT"

import redis

r = redis.Redis (host='localhost')

print("------------------ set / get ------------------")
print(r.set ("name", "DeGizmo"))
print(r.get ("name"))

print("------------------ set / inc / decr / get -----")
print(r.set  ("hit_counter", 1))
print(r.incr ("hit_counter"))
print(r.get  ("hit_counter"))
print(r.decr ("hit_counter"))
print(r.get  ("hit_counter"))

print("------------------ rpush / lrange / ... -------")
print(r.rpush  ("members", "Adam"))
print(r.rpush  ("members", "Bob"))
print(r.rpush  ("members", "Carol"))
print(r.lrange ("members", 0, -1))
print(r.llen   ("members"))
print(r.lindex ("members", 1))

print("------------------ dict set -------------------")

print(r.hset ("key1", None, None, {'11' : 'ONE', '1' : 'one'}))
print(r.hset ("key2", None, None, {'22' : 'TWO', '2' : 'two'}))
print(r.hset ("key3", None, None, {'33' : 'TRE', '3' : 'tre'}))

print("------------------ pipeline dict get ----------")
pipe = r.pipeline()

for key in ['key1', 'key2', 'key3'] :
    pipe.hgetall (key)

for val in pipe.execute ():
    print(val)

print("------------------ list keys ------------------")
print(r.keys ("*"))


print("------------------ lua script -----------------")
lua = """
local value  = redis.call ('GET', KEYS[1])
value        = tonumber   (value)
return value * ARGV[1]"""

multiply = r.register_script (lua)
print(r.set ('foo', 2))
print(multiply (keys=['foo'], args=[5]))

print("-----------------------------------------------")

